# Copyright 2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import http.server
import os
import shutil
import socketserver
import subprocess
import tempfile
import textwrap
import threading
from functools import partial

import portage
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground


class test_sync_zipfile_case(TestCase):
    def test_sync_zipfile(self):
        cpv = "dev-libs/A-0"
        ebuilds = {
            cpv: {"EAPI": "8"},
        }
        etag = "foo"

        server = None
        playground = None
        tmpdir = tempfile.mkdtemp()
        try:

            class Handler(http.server.SimpleHTTPRequestHandler):
                def end_headers(self):
                    self.send_header("etag", etag)
                    super().end_headers()

            server = socketserver.TCPServer(
                ("127.0.0.1", 0),
                partial(Handler, directory=tmpdir),
            )
            threading.Thread(target=server.serve_forever, daemon=True).start()

            playground = ResolverPlayground(
                ebuilds=ebuilds,
            )
            settings = playground.settings

            env = settings.environ()

            repos_conf = textwrap.dedent(
                """
                [test_repo]
                location = %(location)s
                sync-type = zipfile
                sync-uri = %(sync-uri)s
                auto-sync = true
            """
            )

            repo_location = f"{playground.eprefix}/var/repositories/test_repo"

            env["PORTAGE_REPOSITORIES"] = repos_conf % {
                "location": repo_location,
                "sync-uri": "http://{}:{}/test_repo.zip".format(*server.server_address),
            }

            shutil.make_archive(os.path.join(tmpdir, "test_repo"), "zip", repo_location)

            ebuild = playground.trees[playground.eroot]["porttree"].dbapi.findname(cpv)
            self.assertTrue(os.path.exists(ebuild))
            shutil.rmtree(repo_location)
            self.assertFalse(os.path.exists(ebuild))

            result = subprocess.run(
                [
                    "emerge",
                    "--sync",
                ],
                env=env,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
            )
            output = result.stdout.decode(errors="replace")
            try:
                self.assertEqual(result.returncode, os.EX_OK)
            except Exception:
                print(output)
                raise

            repo = settings.repositories["test_repo"]
            sync_mod = portage.sync.module_controller.get_class("zipfile")
            status, repo_revision = sync_mod().retrieve_head(options={"repo": repo})
            self.assertEqual(status, os.EX_OK)
            self.assertEqual(repo_revision, etag)
        finally:
            if server is not None:
                server.shutdown()
            shutil.rmtree(tmpdir)
            playground.cleanup()



# SPDX-License-Identifier: Apache-2.0
# Copyright 2012-2017 The Meson development team

from __future__ import annotations

import re
import subprocess, os.path
import typing as T

from .. import mlog, options
from ..mesonlib import first, MesonException, version_compare
from .compilers import Compiler, clike_debug_args


if T.TYPE_CHECKING:
    from .. import build
    from ..options import MutableKeyedOptionDictType
    from ..dependencies import Dependency
    from ..environment import Environment
    from ..linkers.linkers import DynamicLinker
    from ..mesonlib import MachineChoice

swift_optimization_args: T.Dict[str, T.List[str]] = {
    'plain': [],
    '0': [],
    'g': [],
    '1': ['-O'],
    '2': ['-O'],
    '3': ['-O'],
    's': ['-O'],
}

class SwiftCompiler(Compiler):

    LINKER_PREFIX = ['-Xlinker']
    language = 'swift'
    id = 'llvm'

    def __init__(self, exelist: T.List[str], version: str, for_machine: MachineChoice,
                 env: Environment, full_version: T.Optional[str] = None,
                 linker: T.Optional['DynamicLinker'] = None):
        super().__init__([], exelist, version, for_machine, env,
                         full_version=full_version, linker=linker)
        self.version = version
        if self.info.is_darwin():
            try:
                self.sdk_path = subprocess.check_output(['xcrun', '--show-sdk-path'],
                                                        universal_newlines=True,
                                                        encoding='utf-8', stderr=subprocess.STDOUT).strip()
            except subprocess.CalledProcessError as e:
                mlog.error("Failed to get Xcode SDK path: " + e.output)
                raise MesonException('Xcode license not accepted yet. Run `sudo xcodebuild -license`.')
            except FileNotFoundError:
                mlog.error('xcrun not found. Install Xcode to compile Swift code.')
                raise MesonException('Could not detect Xcode. Please install it to compile Swift code.')

    def get_pic_args(self) -> T.List[str]:
        return []

    def get_pie_args(self) -> T.List[str]:
        return []

    def needs_static_linker(self) -> bool:
        return True

    def get_werror_args(self) -> T.List[str]:
        return ['-warnings-as-errors']

    def get_dependency_gen_args(self, outtarget: str, outfile: str) -> T.List[str]:
        return ['-emit-dependencies']

    def get_dependency_compile_args(self, dep: Dependency) -> T.List[str]:
        args = dep.get_compile_args()
        # Some deps might sneak in a hardcoded path to an older macOS SDK, which can
        # cause compilation errors. Let's replace all .sdk paths with the current one.
        # SwiftPM does it this way: https://github.com/swiftlang/swift-package-manager/pull/6772
        # Not tested on anything else than macOS for now.
        if not self.info.is_darwin():
            return args
        pattern = re.compile(r'.*\/MacOSX[^\/]*\.sdk(\/.*|$)')
        for i, arg in enumerate(args):
            if arg.startswith('-I'):
                match = pattern.match(arg)
                if match:
                    args[i] = '-I' + self.sdk_path + match.group(1)
        return args

    def depfile_for_object(self, objfile: str) -> T.Optional[str]:
        return os.path.splitext(objfile)[0] + '.' + self.get_depfile_suffix()

    def get_depfile_suffix(self) -> str:
        return 'd'

    def get_output_args(self, target: str) -> T.List[str]:
        return ['-o', target]

    def get_header_import_args(self, headername: str) -> T.List[str]:
        return ['-import-objc-header', headername]

    def get_warn_args(self, level: str) -> T.List[str]:
        return []

    def get_std_exe_link_args(self) -> T.List[str]:
        return ['-emit-executable']

    def get_module_args(self, modname: str) -> T.List[str]:
        return ['-module-name', modname]

    def get_mod_gen_args(self) -> T.List[str]:
        return ['-emit-module']

    def get_include_args(self, path: str, is_system: bool) -> T.List[str]:
        return ['-I' + path]

    def get_compile_only_args(self) -> T.List[str]:
        return ['-c']

    def get_options(self) -> MutableKeyedOptionDictType:
        opts = super().get_options()

        key = self.form_compileropt_key('std')
        opts[key] = options.UserComboOption(
            self.make_option_name(key),
            'Swift language version.',
            'none',
            # List them with swiftc -frontend -swift-version ''
            choices=['none', '4', '4.2', '5', '6'])

        return opts

    def get_option_std_args(self, target: build.BuildTarget, subproject: T.Optional[str] = None) -> T.List[str]:
        args: T.List[str] = []

        std = self.get_compileropt_value('std', target, subproject)
        assert isinstance(std, str)

        if std != 'none':
            args += ['-swift-version', std]

        # Pass C compiler -std=... arg to swiftc
        c_langs = ['objc', 'c']
        if target.uses_swift_cpp_interop():
            c_langs = ['objcpp', 'cpp', *c_langs]

        c_lang = first(c_langs, lambda x: x in target.compilers)
        if c_lang is not None:
            cc = target.compilers[c_lang]
            args.extend(arg for c_arg in cc.get_option_std_args(target, subproject) for arg in ['-Xcc', c_arg])

        return args

    def get_working_directory_args(self, path: str) -> T.Optional[T.List[str]]:
        if version_compare(self.version, '<4.2'):
            return None

        return ['-working-directory', path]

    def get_cxx_interoperability_args(self, target: T.Optional[build.BuildTarget] = None) -> T.List[str]:
        if target is not None and not target.uses_swift_cpp_interop():
            return []

        if version_compare(self.version, '<5.9'):
            raise MesonException(f'Compiler {self} does not support C++ interoperability')

        return ['-cxx-interoperability-mode=default']

    def get_library_args(self) -> T.List[str]:
        return ['-parse-as-library']

    def compute_parameters_with_absolute_paths(self, parameter_list: T.List[str],
                                               build_dir: str) -> T.List[str]:
        for idx, i in enumerate(parameter_list):
            if i[:2] == '-I' or i[:2] == '-L':
                parameter_list[idx] = i[:2] + os.path.normpath(os.path.join(build_dir, i[2:]))

        return parameter_list

    def sanity_check(self, work_dir: str) -> None:
        src = 'swifttest.swift'
        source_name = os.path.join(work_dir, src)
        output_name = os.path.join(work_dir, 'swifttest')
        extra_flags: T.List[str] = []
        extra_flags += self.environment.coredata.get_external_args(self.for_machine, self.language)
        if self.is_cross:
            extra_flags += self.get_compile_only_args()
        else:
            extra_flags += self.environment.coredata.get_external_link_args(self.for_machine, self.language)
        with open(source_name, 'w', encoding='utf-8') as ofile:
            ofile.write('''print("Swift compilation is working.")
''')
        pc = subprocess.Popen(self.exelist + extra_flags + ['-emit-executable', '-o', output_name, src], cwd=work_dir)
        pc.wait()
        self.run_sanity_check([output_name], work_dir)

    def get_debug_args(self, is_debug: bool) -> T.List[str]:
        return clike_debug_args[is_debug]

    def get_optimization_args(self, optimization_level: str) -> T.List[str]:
        return swift_optimization_args[optimization_level]



"""Support code for distutils test cases."""

import itertools
import os
import pathlib
import shutil
import sys
import sysconfig
import tempfile
from distutils.core import Distribution

import pytest
from more_itertools import always_iterable


@pytest.mark.usefixtures('distutils_managed_tempdir')
class TempdirManager:
    """
    Mix-in class that handles temporary directories for test cases.
    """

    def mkdtemp(self):
        """Create a temporary directory that will be cleaned up.

        Returns the path of the directory.
        """
        d = tempfile.mkdtemp()
        self.tempdirs.append(d)
        return d

    def write_file(self, path, content='xxx'):
        """Writes a file in the given path.

        path can be a string or a sequence.
        """
        pathlib.Path(*always_iterable(path)).write_text(content, encoding='utf-8')

    def create_dist(self, pkg_name='foo', **kw):
        """Will generate a test environment.

        This function creates:
         - a Distribution instance using keywords
         - a temporary directory with a package structure

        It returns the package directory and the distribution
        instance.
        """
        tmp_dir = self.mkdtemp()
        pkg_dir = os.path.join(tmp_dir, pkg_name)
        os.mkdir(pkg_dir)
        dist = Distribution(attrs=kw)

        return pkg_dir, dist


class DummyCommand:
    """Class to store options for retrieval via set_undefined_options()."""

    def __init__(self, **kwargs):
        vars(self).update(kwargs)

    def ensure_finalized(self):
        pass


def copy_xxmodule_c(directory):
    """Helper for tests that need the xxmodule.c source file.

    Example use:

        def test_compile(self):
            copy_xxmodule_c(self.tmpdir)
            self.assertIn('xxmodule.c', os.listdir(self.tmpdir))

    If the source file can be found, it will be copied to *directory*.  If not,
    the test will be skipped.  Errors during copy are not caught.
    """
    shutil.copy(_get_xxmodule_path(), os.path.join(directory, 'xxmodule.c'))


def _get_xxmodule_path():
    source_name = 'xxmodule.c' if sys.version_info > (3, 9) else 'xxmodule-3.8.c'
    return os.path.join(os.path.dirname(__file__), source_name)


def fixup_build_ext(cmd):
    """Function needed to make build_ext tests pass.

    When Python was built with --enable-shared on Unix, -L. is not enough to
    find libpython<blah>.so, because regrtest runs in a tempdir, not in the
    source directory where the .so lives.

    When Python was built with in debug mode on Windows, build_ext commands
    need their debug attribute set, and it is not done automatically for
    some reason.

    This function handles both of these things.  Example use:

        cmd = build_ext(dist)
        support.fixup_build_ext(cmd)
        cmd.ensure_finalized()

    Unlike most other Unix platforms, Mac OS X embeds absolute paths
    to shared libraries into executables, so the fixup is not needed there.
    """
    if os.name == 'nt':
        cmd.debug = sys.executable.endswith('_d.exe')
    elif sysconfig.get_config_var('Py_ENABLE_SHARED'):
        # To further add to the shared builds fun on Unix, we can't just add
        # library_dirs to the Extension() instance because that doesn't get
        # plumbed through to the final compiler command.
        runshared = sysconfig.get_config_var('RUNSHARED')
        if runshared is None:
            cmd.library_dirs = ['.']
        else:
            if sys.platform == 'darwin':
                cmd.library_dirs = []
            else:
                name, equals, value = runshared.partition('=')
                cmd.library_dirs = [d for d in value.split(os.pathsep) if d]


def combine_markers(cls):
    """
    pytest will honor markers as found on the class, but when
    markers are on multiple subclasses, only one appears. Use
    this decorator to combine those markers.
    """
    cls.pytestmark = [
        mark
        for base in itertools.chain([cls], cls.__bases__)
        for mark in getattr(base, 'pytestmark', [])
    ]
    return cls



# gemato: exceptions
# (c) 2017-2023 Michał Górny
# SPDX-License-Identifier: GPL-2.0-or-later

class GematoException(Exception):
    """
    Base class for gemato exceptions. Makes it easier to catch them all.
    """
    pass


class UnsupportedCompression(GematoException):
    __slots__ = ['suffix']

    def __init__(self, suffix):
        super().__init__(suffix)
        self.suffix = suffix

    def __str__(self):
        return f'Unsupported compression suffix: {self.suffix}'


class UnsupportedHash(GematoException):
    __slots__ = ['hash_name']

    def __init__(self, hash_name):
        super().__init__(hash_name)
        self.hash_name = hash_name

    def __str__(self):
        return f'Unsupported hash name: {self.hash_name}'


class ManifestSyntaxError(GematoException):
    def __init__(self, message):
        super().__init__(message)


class ManifestIncompatibleEntry(GematoException):
    __slots__ = ['e1', 'e2', 'diff']

    def __init__(self, e1, e2, diff):
        super().__init__(e1, e2, diff)
        self.e1 = e1
        self.e2 = e2
        self.diff = diff

    def __str__(self):
        msg = f'Incompatible Manifest entries for {self.e1.path}'
        for k, d1, d2 in self.diff:
            msg += f'\n  {k}: e1: {d1}, e2: {d2}'
        return msg


class ManifestMismatch(GematoException):
    """
    An exception raised for verification failure.
    """

    __slots__ = ['path', 'entry', 'diff']

    def __init__(self, path, entry, diff):
        super().__init__(path, entry, diff)
        self.path = path
        self.entry = entry
        self.diff = diff

    def __str__(self):
        msg = f'Manifest mismatch for {self.path}'
        for k, exp, got in self.diff:
            msg += f'\n  {k}: expected: {exp}, have: {got}'
        return msg


class ManifestCrossDevice(GematoException):
    """
    An exception caused by attempting to cross filesystem boundaries.
    """

    __slots__ = ['path']

    def __init__(self, path):
        super().__init__(path)
        self.path = path

    def __str__(self):
        return (f'Path {self.path} crosses filesystem boundaries, it '
                f'must be IGNORE-d explicitly')


class ManifestSymlinkLoop(GematoException):
    """
    An exception caused by hitting a symlink loop (symlink to itself
    or a parent directory).
    """

    __slots__ = ['path']

    def __init__(self, path):
        super().__init__(path)
        self.path = path

    def __str__(self):
        return (f'Path {self.path} is a symlink to one of its parent '
                f'directories, it must be IGNORE-d explicitly')


class ManifestUnsignedData(GematoException):
    """
    An exception caused by a Manifest file containing non-whitespace
    outside the OpenPGP-signed part.
    """

    def __str__(self):
        return 'Unsigned data found in an OpenPGP signed Manifest'


class OpenPGPRuntimeError(GematoException):
    """
    Base exception class for OpenPGP runtime errors.
    """

    def __init__(self, output):
        super().__init__(output)
        self.output = output


class OpenPGPKeyImportError(OpenPGPRuntimeError):
    """
    An exception raised when key import fails.
    """

    def __str__(self):
        return f'OpenPGP key import failed:\n{self.output}'


class OpenPGPKeyListingError(OpenPGPRuntimeError):
    """
    An exception raised when key listing fails.
    """

    def __str__(self):
        return f'OpenPGP key listing failed:\n{self.output}'


class OpenPGPKeyRefreshError(OpenPGPRuntimeError):
    """
    An exception raised when keyring refresh (update) fails.
    """

    def __str__(self):
        return f'OpenPGP keyring refresh failed:\n{self.output}'


class OpenPGPVerificationFailure(OpenPGPRuntimeError):
    """
    An exception raised when OpenPGP verification fails.
    """

    def __init__(self, output, sig_data=None):
        super().__init__(output)
        self.sig_data = sig_data

    def __str__(self):
        return f'OpenPGP verification failed:\n{self.output}'


class OpenPGPExpiredKeyFailure(OpenPGPVerificationFailure):
    """
    OpenPGP verification rejected because of expired key.
    """

    def __str__(self):
        return (f'OpenPGP signature rejected because of expired key:\n'
                f'{self.output}')


class OpenPGPRevokedKeyFailure(OpenPGPVerificationFailure):
    """
    OpenPGP verification rejected because of revoked key.
    """

    def __str__(self):
        return (f'OpenPGP signature rejected because of revoked key:\n'
                f'{self.output}')


class OpenPGPUnknownSigFailure(OpenPGPVerificationFailure):
    """
    OpenPGP verification rejected for unknown reason (i.e. unrecognized
    GPG status).
    """

    def __str__(self):
        return (f'OpenPGP signature rejected for unknown reason:\n'
                f'{self.output}')


class OpenPGPUntrustedSigFailure(OpenPGPVerificationFailure):
    """OpenPGP verification failed due to untrusted signing key"""

    def __str__(self):
        return (f'Good OpenPGP signature made using untrusted key:\n'
                f'{self.output}')


class OpenPGPSigningFailure(OpenPGPRuntimeError):
    """
    An exception raised when OpenPGP signing fails.
    """

    def __str__(self):
        return f'OpenPGP signing failed:\n{self.output}'


class OpenPGPNoImplementation(GematoException):
    """
    An exception raised when no supported OpenPGP implementation
    is available.
    """

    __slots__ = ['detail']

    def __init__(self, detail):
        super().__init__(detail)
        self.detail = detail

    def __str__(self):
        return f'Requested OpenPGP provider not found ({self.detail})'


class ManifestInvalidPath(GematoException):
    """
    An exception raised when an invalid path tries to be added to
    Manifest.
    """

    __slots__ = ['path', 'detail']

    def __init__(self, path, detail):
        super().__init__(path, detail)
        self.path = path
        self.detail = detail

    def __str__(self):
        return (f'Attempting to add invalid path {self.path} to '
                f'Manifest: {self.detail[0]} must not be '
                f'{self.detail[1]}')


class ManifestInvalidFilename(GematoException):
    """
    An exception raised when an entry for invalid filename is created.
    """

    __slots__ = ['filename', 'pos']

    def __init__(self, filename, pos):
        super().__init__(filename, pos)
        self.filename = filename
        self.pos = pos

    def __str__(self):
        return (f'Attempting to add invalid filename {self.filename!r} '
                f'to Manifest: disallowed character '
                f'U+{ord(self.filename[self.pos]):04X} at position '
                f'{self.pos}')


class ManifestNoSupportedHashes(GematoException):
    """
    An exception raised when all hashes given for an entry are
    unsupported.
    """

    def __init__(self, path, entry):
        super().__init__()
        self.path = path
        self.entry = entry

    def __str__(self):
        return (f"No hashes provided for file {self.path!r} are "
                f"supported (out of {' '.join(self.entry.checksums)})")


class ManifestInsecureHashes(GematoException):
    """
    An exception raised when some of the requested hashes are insecure.
    """

    def __init__(self, hashes):
        super().__init__()
        self.hashes = hashes

    def __str__(self):
        if not self.hashes:
            return "No hashes found when at least one secure hash is required"
        return ("Some of the requested hashes are considered insecure: "
                f"{' '.join(self.hashes)}")



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SlotOperatorUnsolvedTestCase(TestCase):
    """
    Demonstrate bug #456340, where an unsolved circular dependency
    interacts with an unsatisfied built slot-operator dep.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def testSlotOperatorUnsolved(self):
        ebuilds = {
            "dev-libs/icu-50.1.2": {"EAPI": "5", "SLOT": "0/50.1.2"},
            "net-libs/webkit-gtk-1.10.2-r300": {
                "EAPI": "5",
                "DEPEND": ">=dev-libs/icu-3.8.1-r1:=",
                "RDEPEND": ">=dev-libs/icu-3.8.1-r1:=",
            },
            "dev-ruby/rdoc-3.12.1": {
                "EAPI": "5",
                "IUSE": "test",
                "DEPEND": "test? ( >=dev-ruby/hoe-2.7.0 )",
            },
            "dev-ruby/hoe-2.13.0": {
                "EAPI": "5",
                "IUSE": "test",
                "DEPEND": "test? ( >=dev-ruby/rdoc-3.10 )",
                "RDEPEND": "test? ( >=dev-ruby/rdoc-3.10 )",
            },
        }

        binpkgs = {
            "net-libs/webkit-gtk-1.10.2-r300": {
                "EAPI": "5",
                "DEPEND": ">=dev-libs/icu-3.8.1-r1:0/50=",
                "RDEPEND": ">=dev-libs/icu-3.8.1-r1:0/50=",
            },
        }

        installed = {
            "dev-libs/icu-50.1.2": {"EAPI": "5", "SLOT": "0/50.1.2"},
            "net-libs/webkit-gtk-1.10.2-r300": {
                "EAPI": "5",
                "DEPEND": ">=dev-libs/icu-3.8.1-r1:0/50=",
                "RDEPEND": ">=dev-libs/icu-3.8.1-r1:0/50=",
            },
        }

        user_config = {"make.conf": ("FEATURES=test",)}

        world = ["net-libs/webkit-gtk", "dev-ruby/hoe"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True, "--usepkg": True},
                circular_dependency_solutions={
                    "dev-ruby/hoe-2.13.0": frozenset([frozenset([("test", False)])]),
                    "dev-ruby/rdoc-3.12.1": frozenset([frozenset([("test", False)])]),
                },
                success=False,
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                _user_config = user_config.copy()
                _user_config["make.conf"] += (f'BINPKG_FORMAT="{binpkg_format}"',)
                playground = ResolverPlayground(
                    ebuilds=ebuilds,
                    binpkgs=binpkgs,
                    installed=installed,
                    user_config=_user_config,
                    world=world,
                    debug=False,
                )
                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.cleanup()



# Copyright 2018-2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import errno
import os
import pty
import shutil
import socket
import tempfile

from portage.tests import TestCase
from portage.util.futures import asyncio
from portage.util.futures.unix_events import (
    _set_nonblocking,
)


class _PipeClosedTestCase:
    def test_pipe(self):
        read_end, write_end = os.pipe()
        self._do_test(read_end, write_end)

    def test_pty_device(self):
        try:
            read_end, write_end = pty.openpty()
        except OSError:
            self.skipTest("pty not available")
        self._do_test(read_end, write_end)

    def test_domain_socket(self):
        read_end, write_end = socket.socketpair()
        self._do_test(read_end.detach(), write_end.detach())

    def test_named_pipe(self):
        tempdir = tempfile.mkdtemp()
        try:
            fifo_path = os.path.join(tempdir, "fifo")
            os.mkfifo(fifo_path)
            self._do_test(
                os.open(fifo_path, os.O_NONBLOCK | os.O_RDONLY),
                os.open(fifo_path, os.O_NONBLOCK | os.O_WRONLY),
            )
        finally:
            shutil.rmtree(tempdir)


class ReaderPipeClosedTestCase(_PipeClosedTestCase, TestCase):
    """
    Test that a reader callback is called after the other end of
    the pipe has been closed.
    """

    def _do_test(self, read_end, write_end):
        loop = asyncio._wrap_loop()
        read_end = os.fdopen(read_end, "rb", 0)
        write_end = os.fdopen(write_end, "wb", 0)
        try:

            def reader_callback():
                if not reader_callback.called.done():
                    reader_callback.called.set_result(None)

            reader_callback.called = loop.create_future()
            loop.add_reader(read_end.fileno(), reader_callback)

            # Allow the loop to check for IO events, and assert
            # that our future is still not done.
            loop.run_until_complete(asyncio.sleep(0, loop=loop))
            self.assertFalse(reader_callback.called.done())

            # Demonstrate that the callback is called after the
            # other end of the pipe has been closed.
            write_end.close()
            loop.run_until_complete(reader_callback.called)
        finally:
            loop.remove_reader(read_end.fileno())
            write_end.close()
            read_end.close()


class WriterPipeClosedTestCase(_PipeClosedTestCase, TestCase):
    """
    Test that a writer callback is called after the other end of
    the pipe has been closed.
    """

    def _do_test(self, read_end, write_end):
        loop = asyncio._wrap_loop()
        read_end = os.fdopen(read_end, "rb", 0)
        write_end = os.fdopen(write_end, "wb", 0)
        try:

            def writer_callback():
                if not writer_callback.called.done():
                    writer_callback.called.set_result(None)

            writer_callback.called = loop.create_future()
            _set_nonblocking(write_end.fileno())
            loop.add_writer(write_end.fileno(), writer_callback)

            # With pypy we've seen intermittent spurious writer callbacks
            # here, so retry until the correct state is achieved.
            tries = 10
            while tries:
                tries -= 1

                # Fill up the pipe, so that no writer callbacks should be
                # received until the state has changed.
                while True:
                    try:
                        os.write(write_end.fileno(), 512 * b"0")
                    except OSError as e:
                        if e.errno != errno.EAGAIN:
                            raise
                        break

                # Allow the loop to check for IO events, and assert
                # that our future is still not done.
                loop.run_until_complete(asyncio.sleep(0, loop=loop))
                if writer_callback.called.done():
                    writer_callback.called = loop.create_future()
                else:
                    break

            self.assertFalse(writer_callback.called.done())

            # Demonstrate that the callback is called after the
            # other end of the pipe has been closed.
            read_end.close()
            loop.run_until_complete(writer_callback.called)
        finally:
            loop.remove_writer(write_end.fileno())
            write_end.close()
            read_end.close()



from __future__ import annotations

import sys

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import Dict
from typing import TypedDict
from typing import cast

if sys.version_info >= (3, 11):
    from tomllib import loads as load_toml
else:
    from tomli import loads as load_toml

if TYPE_CHECKING:
    if sys.version_info >= (3, 10):
        from typing import TypeAlias
    else:
        from typing_extensions import TypeAlias

from .. import _log

log = _log.log.getChild("toml")

TOML_RESULT: TypeAlias = Dict[str, Any]
TOML_LOADER: TypeAlias = Callable[[str], TOML_RESULT]


class InvalidTomlError(ValueError):
    """Raised when TOML data cannot be parsed."""


def read_toml_content(path: Path, default: TOML_RESULT | None = None) -> TOML_RESULT:
    try:
        data = path.read_text(encoding="utf-8")
    except FileNotFoundError:
        if default is None:
            raise
        else:
            log.debug("%s missing, presuming default %r", path, default)
            return default
    else:
        try:
            return load_toml(data)
        except Exception as e:  # tomllib/tomli raise different decode errors
            raise InvalidTomlError(f"Invalid TOML in {path}") from e


class _CheatTomlData(TypedDict):
    cheat: dict[str, Any]


def load_toml_or_inline_map(data: str | None) -> dict[str, Any]:
    """
    load toml data - with a special hack if only a inline map is given
    """
    if not data:
        return {}
    try:
        if data[0] == "{":
            data = "cheat=" + data
            loaded: _CheatTomlData = cast(_CheatTomlData, load_toml(data))
            return loaded["cheat"]
        return load_toml(data)
    except Exception as e:  # tomllib/tomli raise different decode errors
        raise InvalidTomlError("Invalid TOML content") from e



# SPDX-License-Identifier: Apache-2.0
# Copyright 2019 The Meson development team

from __future__ import annotations

import os
import subprocess
from pathlib import Path
import typing as T

def ls_as_bytestream() -> bytes:
    if os.path.exists('.git'):
        return subprocess.run(['git', 'ls-tree', '-r', '--name-only', 'HEAD'],
                              stdout=subprocess.PIPE).stdout

    files = [str(p) for p in Path('.').glob('**/*')
             if not p.is_dir() and
             not next((x for x in p.parts if x.startswith('.')), None)]
    return '\n'.join(files).encode()


def cscope() -> int:
    ls = b'\n'.join([b'"%s"' % f for f in ls_as_bytestream().split()])
    return subprocess.run(['cscope', '-v', '-b', '-i-'], input=ls).returncode


def ctags() -> int:
    ls = ls_as_bytestream()
    return subprocess.run(['ctags', '-L-'], input=ls).returncode


def etags() -> int:
    ls = ls_as_bytestream()
    return subprocess.run(['etags', '-'], input=ls).returncode


def run(args: T.List[str]) -> int:
    tool_name = args[0]
    srcdir_name = args[1]
    os.chdir(srcdir_name)
    assert tool_name in {'cscope', 'ctags', 'etags'}
    res = globals()[tool_name]()
    assert isinstance(res, int)
    return res



from Cython.Compiler.ModuleNode import ModuleNode
from Cython.Compiler.Symtab import ModuleScope
from Cython.TestUtils import TransformTest
from Cython.Compiler.Visitor import MethodDispatcherTransform
from Cython.Compiler.ParseTreeTransforms import (
    NormalizeTree, AnalyseDeclarationsTransform,
    AnalyseExpressionsTransform, InterpretCompilerDirectives)


class TestMethodDispatcherTransform(TransformTest):
    _tree = None

    def _build_tree(self):
        if self._tree is None:
            context = None

            def fake_module(node):
                scope = ModuleScope('test', None, None)
                return ModuleNode(node.pos, doc=None, body=node,
                                  scope=scope, full_module_name='test',
                                  directive_comments={})
            pipeline = [
                fake_module,
                NormalizeTree(context),
                InterpretCompilerDirectives(context, {}),
                AnalyseDeclarationsTransform(context),
                AnalyseExpressionsTransform(context),
            ]
            self._tree = self.run_pipeline(pipeline, """
                cdef bytes s = b'asdfg'
                cdef dict d = {1:2}
                x = s * 3
                d.get('test')
            """)
        return self._tree

    def test_builtin_method(self):
        calls = [0]
        class Test(MethodDispatcherTransform):
            def _handle_simple_method_dict_get(self, node, func, args, unbound):
                calls[0] += 1
                return node

        tree = self._build_tree()
        Test(None)(tree)
        self.assertEqual(1, calls[0])

    def test_binop_method(self):
        calls = {'bytes': 0, 'object': 0}
        class Test(MethodDispatcherTransform):
            def _handle_simple_method_bytes___mul__(self, node, func, args, unbound):
                calls['bytes'] += 1
                return node
            def _handle_simple_method_object___mul__(self, node, func, args, unbound):
                calls['object'] += 1
                return node

        tree = self._build_tree()
        Test(None)(tree)
        self.assertEqual(1, calls['bytes'])
        self.assertEqual(0, calls['object'])



# SPDX-License-Identifier: Apache-2.0
# Copyright 2015-2016 The Meson development team

from __future__ import annotations

'''CD into dir given as first argument and execute
the command given in the rest of the arguments.'''

import os, subprocess, sys
import typing as T

def run(args: T.List[str]) -> int:
    dirname = args[0]
    command = args[1:]

    os.chdir(dirname)
    return subprocess.call(command)

if __name__ == '__main__':
    sys.exit(run(sys.argv[1:]))



"""
requests.auth
~~~~~~~~~~~~~

This module contains the authentication handlers for Requests.
"""

import hashlib
import os
import re
import threading
import time
import warnings
from base64 import b64encode

from ._internal_utils import to_native_string
from .compat import basestring, str, urlparse
from .cookies import extract_cookies_to_jar
from .utils import parse_dict_header

CONTENT_TYPE_FORM_URLENCODED = "application/x-www-form-urlencoded"
CONTENT_TYPE_MULTI_PART = "multipart/form-data"


def _basic_auth_str(username, password):
    """Returns a Basic Auth string."""

    # "I want us to put a big-ol' comment on top of it that
    # says that this behaviour is dumb but we need to preserve
    # it because people are relying on it."
    #    - Lukasa
    #
    # These are here solely to maintain backwards compatibility
    # for things like ints. This will be removed in 3.0.0.
    if not isinstance(username, basestring):
        warnings.warn(
            "Non-string usernames will no longer be supported in Requests "
            f"3.0.0. Please convert the object you've passed in ({username!r}) to "
            "a string or bytes object in the near future to avoid "
            "problems.",
            category=DeprecationWarning,
        )
        username = str(username)

    if not isinstance(password, basestring):
        warnings.warn(
            "Non-string passwords will no longer be supported in Requests "
            f"3.0.0. Please convert the object you've passed in ({type(password)!r}) to "
            "a string or bytes object in the near future to avoid "
            "problems.",
            category=DeprecationWarning,
        )
        password = str(password)
    # -- End Removal --

    if isinstance(username, str):
        username = username.encode("latin1")

    if isinstance(password, str):
        password = password.encode("latin1")

    authstr = "Basic " + to_native_string(
        b64encode(b":".join((username, password))).strip()
    )

    return authstr


class AuthBase:
    """Base class that all auth implementations derive from"""

    def __call__(self, r):
        raise NotImplementedError("Auth hooks must be callable.")


class HTTPBasicAuth(AuthBase):
    """Attaches HTTP Basic Authentication to the given Request object."""

    def __init__(self, username, password):
        self.username = username
        self.password = password

    def __eq__(self, other):
        return all(
            [
                self.username == getattr(other, "username", None),
                self.password == getattr(other, "password", None),
            ]
        )

    def __ne__(self, other):
        return not self == other

    def __call__(self, r):
        r.headers["Authorization"] = _basic_auth_str(self.username, self.password)
        return r


class HTTPProxyAuth(HTTPBasicAuth):
    """Attaches HTTP Proxy Authentication to a given Request object."""

    def __call__(self, r):
        r.headers["Proxy-Authorization"] = _basic_auth_str(self.username, self.password)
        return r


class HTTPDigestAuth(AuthBase):
    """Attaches HTTP Digest Authentication to the given Request object."""

    def __init__(self, username, password):
        self.username = username
        self.password = password
        # Keep state in per-thread local storage
        self._thread_local = threading.local()

    def init_per_thread_state(self):
        # Ensure state is initialized just once per-thread
        if not hasattr(self._thread_local, "init"):
            self._thread_local.init = True
            self._thread_local.last_nonce = ""
            self._thread_local.nonce_count = 0
            self._thread_local.chal = {}
            self._thread_local.pos = None
            self._thread_local.num_401_calls = None

    def build_digest_header(self, method, url):
        """
        :rtype: str
        """

        realm = self._thread_local.chal["realm"]
        nonce = self._thread_local.chal["nonce"]
        qop = self._thread_local.chal.get("qop")
        algorithm = self._thread_local.chal.get("algorithm")
        opaque = self._thread_local.chal.get("opaque")
        hash_utf8 = None

        if algorithm is None:
            _algorithm = "MD5"
        else:
            _algorithm = algorithm.upper()
        # lambdas assume digest modules are imported at the top level
        if _algorithm == "MD5" or _algorithm == "MD5-SESS":

            def md5_utf8(x):
                if isinstance(x, str):
                    x = x.encode("utf-8")
                return hashlib.md5(x).hexdigest()

            hash_utf8 = md5_utf8
        elif _algorithm == "SHA":

            def sha_utf8(x):
                if isinstance(x, str):
                    x = x.encode("utf-8")
                return hashlib.sha1(x).hexdigest()

            hash_utf8 = sha_utf8
        elif _algorithm == "SHA-256":

            def sha256_utf8(x):
                if isinstance(x, str):
                    x = x.encode("utf-8")
                return hashlib.sha256(x).hexdigest()

            hash_utf8 = sha256_utf8
        elif _algorithm == "SHA-512":

            def sha512_utf8(x):
                if isinstance(x, str):
                    x = x.encode("utf-8")
                return hashlib.sha512(x).hexdigest()

            hash_utf8 = sha512_utf8

        KD = lambda s, d: hash_utf8(f"{s}:{d}")  # noqa:E731

        if hash_utf8 is None:
            return None

        # XXX not implemented yet
        entdig = None
        p_parsed = urlparse(url)
        #: path is request-uri defined in RFC 2616 which should not be empty
        path = p_parsed.path or "/"
        if p_parsed.query:
            path += f"?{p_parsed.query}"

        A1 = f"{self.username}:{realm}:{self.password}"
        A2 = f"{method}:{path}"

        HA1 = hash_utf8(A1)
        HA2 = hash_utf8(A2)

        if nonce == self._thread_local.last_nonce:
            self._thread_local.nonce_count += 1
        else:
            self._thread_local.nonce_count = 1
        ncvalue = f"{self._thread_local.nonce_count:08x}"
        s = str(self._thread_local.nonce_count).encode("utf-8")
        s += nonce.encode("utf-8")
        s += time.ctime().encode("utf-8")
        s += os.urandom(8)

        cnonce = hashlib.sha1(s).hexdigest()[:16]
        if _algorithm == "MD5-SESS":
            HA1 = hash_utf8(f"{HA1}:{nonce}:{cnonce}")

        if not qop:
            respdig = KD(HA1, f"{nonce}:{HA2}")
        elif qop == "auth" or "auth" in qop.split(","):
            noncebit = f"{nonce}:{ncvalue}:{cnonce}:auth:{HA2}"
            respdig = KD(HA1, noncebit)
        else:
            # XXX handle auth-int.
            return None

        self._thread_local.last_nonce = nonce

        # XXX should the partial digests be encoded too?
        base = (
            f'username="{self.username}", realm="{realm}", nonce="{nonce}", '
            f'uri="{path}", response="{respdig}"'
        )
        if opaque:
            base += f', opaque="{opaque}"'
        if algorithm:
            base += f', algorithm="{algorithm}"'
        if entdig:
            base += f', digest="{entdig}"'
        if qop:
            base += f', qop="auth", nc={ncvalue}, cnonce="{cnonce}"'

        return f"Digest {base}"

    def handle_redirect(self, r, **kwargs):
        """Reset num_401_calls counter on redirects."""
        if r.is_redirect:
            self._thread_local.num_401_calls = 1

    def handle_401(self, r, **kwargs):
        """
        Takes the given response and tries digest-auth, if needed.

        :rtype: requests.Response
        """

        # If response is not 4xx, do not auth
        # See https://github.com/psf/requests/issues/3772
        if not 400 <= r.status_code < 500:
            self._thread_local.num_401_calls = 1
            return r

        if self._thread_local.pos is not None:
            # Rewind the file position indicator of the body to where
            # it was to resend the request.
            r.request.body.seek(self._thread_local.pos)
        s_auth = r.headers.get("www-authenticate", "")

        if "digest" in s_auth.lower() and self._thread_local.num_401_calls < 2:
            self._thread_local.num_401_calls += 1
            pat = re.compile(r"digest ", flags=re.IGNORECASE)
            self._thread_local.chal = parse_dict_header(pat.sub("", s_auth, count=1))

            # Consume content and release the original connection
            # to allow our new request to reuse the same one.
            r.content
            r.close()
            prep = r.request.copy()
            extract_cookies_to_jar(prep._cookies, r.request, r.raw)
            prep.prepare_cookies(prep._cookies)

            prep.headers["Authorization"] = self.build_digest_header(
                prep.method, prep.url
            )
            _r = r.connection.send(prep, **kwargs)
            _r.history.append(r)
            _r.request = prep

            return _r

        self._thread_local.num_401_calls = 1
        return r

    def __call__(self, r):
        # Initialize per-thread state, if needed
        self.init_per_thread_state()
        # If we have a saved nonce, skip the 401
        if self._thread_local.last_nonce:
            r.headers["Authorization"] = self.build_digest_header(r.method, r.url)
        try:
            self._thread_local.pos = r.body.tell()
        except AttributeError:
            # In the case of HTTPDigestAuth being reused and the body of
            # the previous request was a file-like object, pos has the
            # file position of the previous body. Ensure it's set to
            # None.
            self._thread_local.pos = None
        r.register_hook("response", self.handle_401)
        r.register_hook("response", self.handle_redirect)
        self._thread_local.num_401_calls = 1

        return r

    def __eq__(self, other):
        return all(
            [
                self.username == getattr(other, "username", None),
                self.password == getattr(other, "password", None),
            ]
        )

    def __ne__(self, other):
        return not self == other



# SPDX-License-Identifier: Apache-2.0
# Copyright 2019 The Meson development team

from __future__ import annotations

import os
import argparse
import subprocess
from pathlib import Path
import typing as T

from ..mesonlib import Popen_safe, split_args, determine_worker_count
from .. import mlog

class ExternalProject:
    def __init__(self, options: argparse.Namespace):
        self.name = options.name
        self.src_dir = options.srcdir
        self.build_dir = options.builddir
        self.install_dir = options.installdir
        self.log_dir = options.logdir
        self.verbose = options.verbose
        self.stampfile = options.stampfile
        self.depfile = options.depfile
        self.make = split_args(options.make)

    def write_depfile(self) -> None:
        with open(self.depfile, 'w', encoding='utf-8') as f:
            f.write(f'{self.stampfile}: \\\n')
            for dirpath, dirnames, filenames in os.walk(self.src_dir):
                dirnames[:] = [d for d in dirnames if not d.startswith('.')]
                for fname in filenames:
                    if fname.startswith('.'):
                        continue
                    path = Path(dirpath, fname)
                    f.write('  {} \\\n'.format(path.as_posix().replace(' ', '\\ ')))

    def write_stampfile(self) -> None:
        with open(self.stampfile, 'w', encoding='utf-8'):
            pass

    def supports_jobs_flag(self) -> bool:
        p, o, e = Popen_safe(self.make + ['--version'])
        if p.returncode == 0 and ('GNU Make' in o or 'waf' in o):
            return True
        return False

    def build(self) -> int:
        make_cmd = self.make.copy()
        if self.supports_jobs_flag():
            make_cmd.append(f'-j{determine_worker_count()}')
        rc = self._run('build', make_cmd)
        if rc != 0:
            return rc

        install_cmd = self.make.copy()
        install_env = {}
        install_env['DESTDIR'] = self.install_dir
        install_cmd.append('install')
        rc = self._run('install', install_cmd, install_env)
        if rc != 0:
            return rc

        self.write_depfile()
        self.write_stampfile()

        return 0

    def _run(self, step: str, command: T.List[str], env: T.Optional[T.Dict[str, str]] = None) -> int:
        m = 'Running command ' + str(command) + ' in directory ' + str(self.build_dir) + '\n'
        logfile = Path(self.log_dir, f'{self.name}-{step}.log')
        output = None
        if not self.verbose:
            output = open(logfile, 'w', encoding='utf-8')
            output.write(m + '\n')
            output.flush()
        else:
            print(m)
        run_env = os.environ.copy()
        if env:
            run_env.update(env)
        p, o, e = Popen_safe(command, stderr=subprocess.STDOUT, stdout=output,
                             cwd=self.build_dir,
                             env=run_env)
        if p.returncode != 0:
            m = f'{step} step returned error code {p.returncode}.'
            if not self.verbose:
                m += '\nSee logs: ' + str(logfile)
            print(m)
            contents = mlog.ci_fold_file(logfile, f'CI platform detected, click here for {os.path.basename(logfile)} contents.')
            if contents:
                print(contents)
        return p.returncode

def run(args: T.List[str]) -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument('--name')
    parser.add_argument('--srcdir')
    parser.add_argument('--builddir')
    parser.add_argument('--installdir')
    parser.add_argument('--logdir')
    parser.add_argument('--make')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('stampfile')
    parser.add_argument('depfile')

    options = parser.parse_args(args)
    ep = ExternalProject(options)
    return ep.build()



# Copyright 2012-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import threading

from portage import os
from _emerge.AbstractPollTask import AbstractPollTask


class PipeReaderBlockingIO(AbstractPollTask):
    """
    Reads output from one or more files and saves it in memory, for
    retrieval via the getvalue() method. This is driven by a thread
    for each input file, in order to support blocking IO.  This is
    historically useful for using threads to handle blocking IO with
    Jython, since Jython lacks the fcntl module which is needed for
    non-blocking IO (see http://bugs.jython.org/issue1074).

    Portage does not currently support Jython, but re-introducing
    support in The Future (TM) may be possible.
    """

    __slots__ = ("input_files", "_read_data", "_terminate", "_threads", "_thread_rlock")

    def _start(self):
        self._terminate = threading.Event()
        self._threads = {}
        self._read_data = []

        self._registered = True
        self._thread_rlock = threading.RLock()
        with self._thread_rlock:
            for f in self.input_files.values():
                t = threading.Thread(target=self._reader_thread, args=(f,))
                t.daemon = True
                t.start()
                self._threads[f] = t

    def _reader_thread(self, f):
        terminated = self._terminate.is_set
        bufsize = self._bufsize
        while not terminated():
            buf = f.read(bufsize)
            with self._thread_rlock:
                if terminated():
                    break
                elif buf:
                    self._read_data.append(buf)
                else:
                    del self._threads[f]
                    if not self._threads:
                        # Thread-safe callback to EventLoop
                        self.scheduler.call_soon_threadsafe(self._eof)
                    break
        f.close()

    def _eof(self):
        self._registered = False
        if self.returncode is None:
            self.returncode = os.EX_OK
        self._async_wait()

    def _cancel(self):
        self._terminate.set()
        self._registered = False
        if self.returncode is None:
            self.returncode = self._cancelled_returncode
        self._async_wait()

    def getvalue(self):
        """Retrieve the entire contents"""
        with self._thread_rlock:
            return b"".join(self._read_data)

    def close(self):
        """Free the memory buffer."""
        with self._thread_rlock:
            self._read_data = None



# SPDX-License-Identifier: Apache-2.0
# Copyright 2012-2022 The Meson development team

from __future__ import annotations

from ..mesonlib import (
    MesonException, EnvironmentException, MachineChoice, join_args,
    search_version, is_windows, Popen_safe, Popen_safe_logged, version_compare, windows_proof_rm,
)
from ..programs import ExternalProgram
from ..envconfig import BinaryTable, detect_cpu_family
from .. import mlog

from ..linkers import guess_win_linker, guess_nix_linker

import subprocess
import platform
import re
import shutil
import tempfile
import os
import typing as T

if T.TYPE_CHECKING:
    from .compilers import Compiler
    from .asm import ASMCompiler
    from .c import CCompiler
    from .cpp import CPPCompiler
    from .fortran import FortranCompiler
    from .rust import RustCompiler
    from ..linkers.linkers import StaticLinker, DynamicLinker
    from ..environment import Environment


# Default compilers and linkers
# =============================

defaults: T.Dict[str, T.List[str]] = {}

# List of potential compilers.
if is_windows():
    # Intel C and C++ compiler is icl on Windows, but icc and icpc elsewhere.
    # Search for icl before cl, since Intel "helpfully" provides a
    # cl.exe that returns *exactly the same thing* that Microsoft's
    # cl.exe does, and if icl is present, it's almost certainly what
    # you want.
    defaults['c'] = ['icl', 'cl', 'cc', 'gcc', 'clang', 'clang-cl', 'pgcc']
    # There is currently no pgc++ for Windows, only for  Mac and Linux.
    defaults['cpp'] = ['icl', 'cl', 'c++', 'g++', 'clang++', 'clang-cl']
    # the binary flang-new will be renamed to flang in the foreseeable future
    defaults['fortran'] = ['ifort', 'ifx', 'gfortran', 'flang-new', 'flang', 'pgfortran', 'g95']
    defaults['objc'] = ['clang', 'clang-cl', 'gcc']
    defaults['objcpp'] = ['clang++', 'clang-cl', 'g++']
    defaults['cs'] = ['csc', 'mcs']
else:
    if platform.machine().lower() == 'e2k':
        defaults['c'] = ['cc', 'gcc', 'lcc', 'clang']
        defaults['cpp'] = ['c++', 'g++', 'l++', 'clang++']
        defaults['objc'] = ['clang']
        defaults['objcpp'] = ['clang++']
    else:
        defaults['c'] = ['cc', 'gcc', 'clang', 'nvc', 'pgcc', 'icc', 'icx']
        defaults['cpp'] = ['c++', 'g++', 'clang++', 'nvc++', 'pgc++', 'icpc', 'icpx']
        defaults['objc'] = ['clang', 'gcc']
        defaults['objcpp'] = ['clang++', 'g++']
    # the binary flang-new will be renamed to flang in the foreseeable future
    defaults['fortran'] = ['gfortran', 'flang-new', 'flang', 'nvfortran', 'pgfortran', 'ifort', 'ifx', 'g95']
    defaults['cs'] = ['mcs', 'csc']
defaults['d'] = ['ldc2', 'ldc', 'gdc', 'dmd']
defaults['java'] = ['javac']
defaults['cuda'] = ['nvcc']
defaults['rust'] = ['rustc']
defaults['swift'] = ['swiftc']
defaults['vala'] = ['valac']
defaults['cython'] = ['cython', 'cython3'] # Official name is cython, but Debian renamed it to cython3.
defaults['static_linker'] = ['ar', 'gar']
defaults['strip'] = ['strip']
defaults['vs_static_linker'] = ['lib']
defaults['clang_cl_static_linker'] = ['llvm-lib']
defaults['cuda_static_linker'] = ['nvlink']
defaults['gcc_static_linker'] = ['gcc-ar']
defaults['clang_static_linker'] = ['llvm-ar']
defaults['emxomf_static_linker'] = ['emxomfar']
defaults['nasm'] = ['nasm', 'yasm']


def compiler_from_language(env: 'Environment', lang: str, for_machine: MachineChoice) -> T.Optional[Compiler]:
    lang_map: T.Dict[str, T.Callable[['Environment', MachineChoice], Compiler]] = {
        'c': detect_c_compiler,
        'cpp': detect_cpp_compiler,
        'objc': detect_objc_compiler,
        'cuda': detect_cuda_compiler,
        'objcpp': detect_objcpp_compiler,
        'java': detect_java_compiler,
        'cs': detect_cs_compiler,
        'vala': detect_vala_compiler,
        'd': detect_d_compiler,
        'rust': detect_rust_compiler,
        'fortran': detect_fortran_compiler,
        'swift': detect_swift_compiler,
        'cython': detect_cython_compiler,
        'nasm': detect_nasm_compiler,
        'masm': detect_masm_compiler,
        'linearasm': detect_linearasm_compiler,
    }
    return lang_map[lang](env, for_machine) if lang in lang_map else None

def detect_compiler_for(env: 'Environment', lang: str, for_machine: MachineChoice, skip_sanity_check: bool, subproject: str) -> T.Optional[Compiler]:
    comp = compiler_from_language(env, lang, for_machine)
    if comp is None:
        return comp
    assert comp.for_machine == for_machine
    env.coredata.process_compiler_options(lang, comp, subproject)
    if not skip_sanity_check:
        comp.sanity_check(env.get_scratch_dir())
    env.coredata.compilers[comp.for_machine][lang] = comp
    return comp


# Helpers
# =======

def _get_compilers(env: 'Environment', lang: str, for_machine: MachineChoice,
                   allow_build_machine: bool = False) -> T.Tuple[T.List[T.List[str]], T.Union[None, ExternalProgram]]:
    '''
    The list of compilers is detected in the exact same way for
    C, C++, ObjC, ObjC++, Fortran, CS so consolidate it here.
    '''
    value = env.lookup_binary_entry(for_machine, lang)
    if value is not None:
        comp, ccache = BinaryTable.parse_entry(value)
        # Return value has to be a list of compiler 'choices'
        compilers = [comp]
    else:
        if not env.machines.matches_build_machine(for_machine):
            if allow_build_machine:
                return _get_compilers(env, lang, MachineChoice.BUILD)
            raise EnvironmentException(f'{lang!r} compiler binary not defined in cross file [binaries] section')
        compilers = [[x] for x in defaults[lang]]
        ccache = BinaryTable.detect_compiler_cache()

    return compilers, ccache

def _handle_exceptions(
        exceptions: T.Mapping[str, T.Union[Exception, str]],
        binaries: T.List[T.List[str]],
        bintype: str = 'compiler') -> T.NoReturn:
    errmsg = f'Unknown {bintype}(s): {binaries}'
    if exceptions:
        errmsg += '\nThe following exception(s) were encountered:'
        for c, e in exceptions.items():
            if isinstance(e, MesonException):
                errmsg += f'\nUsing `{c}` failed: {e}'
            else:
                errmsg += f'\nRunning `{c}` gave "{e}"'
    raise EnvironmentException(errmsg)


# Linker specific
# ===============

def detect_static_linker(env: 'Environment', compiler: Compiler) -> StaticLinker:
    from . import d
    from ..linkers import linkers
    from ..options import OptionKey
    linker = env.lookup_binary_entry(compiler.for_machine, 'ar')
    if linker is not None:
        trials = [linker]
    else:
        default_linkers = [[l] for l in defaults['static_linker']]
        if compiler.language == 'cuda':
            trials = [defaults['cuda_static_linker']] + default_linkers
        elif compiler.get_argument_syntax() == 'msvc':
            trials = [defaults['vs_static_linker'], defaults['clang_cl_static_linker']]
        elif env.machines[compiler.for_machine].is_os2() and env.coredata.optstore.get_value_for(OptionKey('os2_emxomf')):
            trials = [defaults['emxomf_static_linker']] + default_linkers
        elif compiler.id == 'gcc':
            # Use gcc-ar if available; needed for LTO
            trials = [defaults['gcc_static_linker']] + default_linkers
        elif compiler.id == 'clang':
            # Use llvm-ar if available; needed for LTO
            llvm_ar = defaults['clang_static_linker']
            # Extract the version major of the compiler to use as a suffix
            suffix = compiler.version.split('.')[0]
            # Prefer suffixed llvm-ar first, then unsuffixed then the defaults
            trials = [[f'{llvm_ar[0]}-{suffix}'], llvm_ar] + default_linkers
        elif compiler.language == 'd':
            # Prefer static linkers over linkers used by D compilers
            if is_windows():
                trials = [defaults['vs_static_linker'], defaults['clang_cl_static_linker'], compiler.get_linker_exelist()]
            else:
                trials = default_linkers
        elif compiler.id == 'intel-cl' and compiler.language == 'c': # why not cpp? Is this a bug?
            # Intel has its own linker that acts like Microsoft's lib
            trials = [['xilib']]
        elif is_windows() and compiler.id == 'pgi': # this handles cpp / nvidia HPC, in addition to just c/fortran
            trials = [['ar']]  # For PGI on Windows, "ar" is just a wrapper calling link/lib.
        elif is_windows() and compiler.id == 'nasm':
            # This may well be LINK.EXE if it's under a MSVC environment
            trials = [defaults['vs_static_linker'], defaults['clang_cl_static_linker']] + default_linkers
        else:
            trials = default_linkers
    popen_exceptions = {}
    for linker in trials:
        linker_name = os.path.basename(linker[0])

        if any(os.path.basename(x) in {'lib', 'lib.exe', 'llvm-lib', 'llvm-lib.exe', 'xilib', 'xilib.exe'} for x in linker):
            arg = '/?'
        elif linker_name in {'ar2000', 'ar2000.exe', 'ar430', 'ar430.exe', 'ar6x', 'ar6x.exe'}:
            arg = '?'
        elif linker_name in {'armar', 'armar.exe'}:
            arg = '-h'
        else:
            arg = '--version'
        try:
            p, out, err = Popen_safe_logged(linker + [arg], msg='Detecting archiver via')
        except OSError as e:
            popen_exceptions[join_args(linker + [arg])] = e
            continue
        if "xilib: executing 'lib'" in err:
            return linkers.IntelVisualStudioLinker(linker, env, getattr(compiler, 'machine', None))
        if '/OUT:' in out.upper() or '/OUT:' in err.upper():
            return linkers.VisualStudioLinker(linker, env, getattr(compiler, 'machine', None))
        if 'ar-Error-Unknown switch: --version' in err:
            return linkers.PGIStaticLinker(linker, env)
        if p.returncode == 0 and 'armar' in linker_name:
            return linkers.ArmarLinker(linker, env)
        if 'DMD32 D Compiler' in out or 'DMD64 D Compiler' in out:
            assert isinstance(compiler, d.DCompiler)
            return linkers.DLinker(linker, env, compiler.arch)
        if 'LDC - the LLVM D compiler' in out:
            assert isinstance(compiler, d.DCompiler)
            return linkers.DLinker(linker, env, compiler.arch, rsp_syntax=compiler.rsp_file_syntax())
        if 'GDC' in out and ' based on D ' in out:
            assert isinstance(compiler, d.DCompiler)
            return linkers.DLinker(linker, env, compiler.arch)
        if err.startswith('Renesas') and 'rlink' in linker_name:
            return linkers.CcrxLinker(linker, env)
        if out.startswith('GNU ar'):
            if 'xc16-ar' in linker_name:
                return linkers.Xc16Linker(linker, env)
            elif 'xc32-ar' in linker_name:
                return linkers.Xc32ArLinker(compiler.for_machine, linker, env)
        if 'Texas Instruments Incorporated' in out:
            if 'ar2000' in linker_name:
                return linkers.C2000Linker(linker, env)
            elif 'ar6000' in linker_name:
                return linkers.C6000Linker(linker, env)
            else:
                return linkers.TILinker(linker, env)
        if out.startswith('The CompCert'):
            return linkers.CompCertLinker(linker, env)
        if out.strip().startswith('Metrowerks') or out.strip().startswith('Freescale'):
            if 'ARM' in out:
                return linkers.MetrowerksStaticLinkerARM(linker, env)
            else:
                return linkers.MetrowerksStaticLinkerEmbeddedPowerPC(linker, env)
        if 'TASKING VX-toolset' in err:
            return linkers.TaskingStaticLinker(linker, env)
        if p.returncode == 0:
            return linkers.ArLinker(compiler.for_machine, linker, env)
        if p.returncode == 1 and err.startswith('usage'): # OSX
            return linkers.AppleArLinker(compiler.for_machine, linker, env)
        if p.returncode == 1 and err.startswith('Usage'): # AIX
            return linkers.AIXArLinker(linker, env)
        if p.returncode == 1 and err.startswith('ar: bad option: --'): # Solaris
            return linkers.ArLinker(compiler.for_machine, linker, env)
        if p.returncode == 1 and err.startswith('emxomfar'):
            return linkers.EmxomfArLinker(compiler.for_machine, linker, env)
    _handle_exceptions(popen_exceptions, trials, 'linker')
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')


# Compilers
# =========


def _detect_c_or_cpp_compiler(env: 'Environment', lang: str, for_machine: MachineChoice, *, override_compiler: T.Optional[T.List[str]] = None) -> Compiler:
    """Shared implementation for finding the C or C++ compiler to use.

    the override_compiler option is provided to allow compilers which use
    the compiler (GCC or Clang usually) as their shared linker, to find
    the linker they need.
    """
    from . import c, cpp
    from ..linkers import linkers
    popen_exceptions: T.Dict[str, T.Union[Exception, str]] = {}
    compilers, ccache_exe = _get_compilers(env, lang, for_machine)
    ccache = ccache_exe.get_command() if (ccache_exe and ccache_exe.found()) else []
    if override_compiler is not None:
        compilers = [override_compiler]
    cls: T.Union[T.Type[CCompiler], T.Type[CPPCompiler]]
    lnk: T.Union[T.Type[StaticLinker], T.Type[DynamicLinker]]

    for compiler in compilers:
        if isinstance(compiler, str):
            compiler = [compiler]
        compiler_name = os.path.basename(compiler[0])

        if any(os.path.basename(x) in {'cl', 'cl.exe', 'clang-cl', 'clang-cl.exe'} for x in compiler):
            # Watcom C provides its own cl.exe clone that mimics an older
            # version of Microsoft's compiler. Since Watcom's cl.exe is
            # just a wrapper, we skip using it if we detect its presence
            # so as not to confuse Meson when configuring for MSVC.
            #
            # Additionally the help text of Watcom's cl.exe is paged, and
            # the binary will not exit without human intervention. In
            # practice, Meson will block waiting for Watcom's cl.exe to
            # exit, which requires user input and thus will never exit.
            if 'WATCOM' in os.environ:
                def sanitize(p: T.Optional[str]) -> T.Optional[str]:
                    return os.path.normcase(os.path.abspath(p)) if p else None

                watcom_cls = [sanitize(os.path.join(os.environ['WATCOM'], 'BINNT', 'cl')),
                              sanitize(os.path.join(os.environ['WATCOM'], 'BINNT', 'cl.exe')),
                              sanitize(os.path.join(os.environ['WATCOM'], 'BINNT64', 'cl')),
                              sanitize(os.path.join(os.environ['WATCOM'], 'BINNT64', 'cl.exe'))]
                found_cl = sanitize(shutil.which('cl'))
                if found_cl in watcom_cls:
                    mlog.debug('Skipping unsupported cl.exe clone at:', found_cl)
                    continue
            arg = '/?'
        elif 'armcc' in compiler_name:
            arg = '--vsn'
        elif 'ccrx' in compiler_name:
            arg = '-v'
        elif 'xc16' in compiler_name:
            arg = '--version'
        elif 'ccomp' in compiler_name:
            arg = '-version'
        elif compiler_name in {'cl2000', 'cl2000.exe', 'cl430', 'cl430.exe', 'armcl', 'armcl.exe', 'cl6x', 'cl6x.exe'}:
            # TI compiler
            arg = '-version'
        elif compiler_name in {'icl', 'icl.exe'}:
            # if you pass anything to icl you get stuck in a pager
            arg = ''
        else:
            arg = '--version'

        cmd = compiler + [arg]
        try:
            p, out, err = Popen_safe_logged(cmd, msg='Detecting compiler via')
        except OSError as e:
            popen_exceptions[join_args(cmd)] = e
            continue

        if 'ccrx' in compiler_name:
            out = err

        full_version = out.split('\n', 1)[0]
        version = search_version(out)

        guess_gcc_or_lcc: T.Optional[str] = None
        if 'Free Software Foundation' in out or out.startswith('xt-'):
            guess_gcc_or_lcc = 'gcc'
        if 'e2k' in out and 'lcc' in out:
            guess_gcc_or_lcc = 'lcc'
        if 'Microchip' in out:
            # this output has "Free Software Foundation" in its version
            guess_gcc_or_lcc = None

        if guess_gcc_or_lcc:
            defines = _get_gnu_compiler_defines(compiler, lang)
            if not defines:
                popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                continue

            if guess_gcc_or_lcc == 'lcc':
                version = _get_lcc_version_from_defines(defines)
                cls = c.ElbrusCCompiler if lang == 'c' else cpp.ElbrusCPPCompiler
            else:
                version = _get_gnu_version_from_defines(defines)
                cls = c.GnuCCompiler if lang == 'c' else cpp.GnuCPPCompiler

            linker = guess_nix_linker(env, compiler, cls, version, for_machine)

            return cls(
                ccache, compiler, version, for_machine,
                env, defines=defines, full_version=full_version,
                linker=linker)

        if 'Emscripten' in out:
            cls = c.EmscriptenCCompiler if lang == 'c' else cpp.EmscriptenCPPCompiler
            env.add_lang_args(cls.language, cls, for_machine)

            # emcc requires a file input in order to pass arguments to the
            # linker. It'll exit with an error code, but still print the
            # linker version.
            with tempfile.NamedTemporaryFile(suffix='.c') as f:
                cmd = compiler + [cls.LINKER_PREFIX + "--version", f.name]
                _, o, _ = Popen_safe(cmd)

            linker = linkers.WASMDynamicLinker(
                compiler, env, for_machine, cls.LINKER_PREFIX,
                [], version=search_version(o))
            return cls(
                ccache, compiler, version, for_machine, env,
                linker=linker, full_version=full_version)

        if 'Arm C/C++/Fortran Compiler' in out:
            arm_ver_match = re.search(r'version (\d+)\.(\d+)\.?(\d+)? \(build number (\d+)\)', out)
            assert arm_ver_match is not None, 'for mypy'  # because mypy *should* be complaining that this could be None
            version = '.'.join([x for x in arm_ver_match.groups() if x is not None])
            if lang == 'c':
                cls = c.ArmLtdClangCCompiler
            elif lang == 'cpp':
                cls = cpp.ArmLtdClangCPPCompiler
            linker = guess_nix_linker(env, compiler, cls, version, for_machine)
            return cls(
                ccache, compiler, version, for_machine, env,
                linker=linker)
        if 'armclang' in out:
            # The compiler version is not present in the first line of output,
            # instead it is present in second line, startswith 'Component:'.
            # So, searching for the 'Component' in out although we know it is
            # present in second line, as we are not sure about the
            # output format in future versions
            arm_ver_match = re.search('.*Component.*', out)
            if arm_ver_match is None:
                popen_exceptions[join_args(compiler)] = 'version string not found'
                continue
            arm_ver_str = arm_ver_match.group(0)
            # Override previous values
            version = search_version(arm_ver_str)
            full_version = arm_ver_str
            cls = c.ArmclangCCompiler if lang == 'c' else cpp.ArmclangCPPCompiler
            linker = linkers.ArmClangDynamicLinker(env, for_machine, version=version)
            env.add_lang_args(cls.language, cls, for_machine)
            return cls(
                ccache, compiler, version, for_machine, env,
                full_version=full_version, linker=linker)
        if 'CL.EXE COMPATIBILITY' in out:
            # if this is clang-cl masquerading as cl, detect it as cl, not
            # clang
            arg = '--version'
            try:
                p, out, err = Popen_safe(compiler + [arg])
            except OSError as e:
                popen_exceptions[join_args(compiler + [arg])] = e
            version = search_version(out)
            match = re.search('^Target: (.*?)-', out, re.MULTILINE)
            if match:
                target = match.group(1)
            else:
                target = 'unknown target'
            cls = c.ClangClCCompiler if lang == 'c' else cpp.ClangClCPPCompiler
            linker = guess_win_linker(env, ['lld-link'], cls, version, for_machine)
            return cls(
                compiler, version, for_machine, env, target,
                linker=linker)

        # must be detected here before clang because TI compilers contain 'clang' in their output and so that they can be detected as 'clang'
        ti_compilers = {
           'TMS320C2000 C/C++': (c.C2000CCompiler, cpp.C2000CPPCompiler, linkers.C2000DynamicLinker),
           'TMS320C6x C/C++': (c.C6000CCompiler, cpp.C6000CPPCompiler, linkers.C6000DynamicLinker),
           'TI ARM C/C++ Compiler': (c.TICCompiler, cpp.TICPPCompiler, linkers.TIDynamicLinker),
           'MSP430 C/C++': (c.TICCompiler, cpp.TICPPCompiler, linkers.TIDynamicLinker)
        }
        for identifier, compiler_classes in ti_compilers.items():
            if identifier in out:
                cls = compiler_classes[0] if lang == 'c' else compiler_classes[1]
                lnk = compiler_classes[2]
                env.add_lang_args(cls.language, cls, for_machine)
                linker = lnk(compiler, env, for_machine, version=version)
                return cls(
                    ccache, compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

        if 'clang' in out or 'Clang' in out:
            linker = None

            defines = _get_clang_compiler_defines(compiler, lang)

            # Even if the for_machine is darwin, we could be using vanilla
            # clang.
            if 'Apple' in out:
                cls = c.AppleClangCCompiler if lang == 'c' else cpp.AppleClangCPPCompiler
            else:
                cls = c.ClangCCompiler if lang == 'c' else cpp.ClangCPPCompiler

            if 'windows' in out or env.machines[for_machine].is_windows():
                # If we're in a MINGW context this actually will use a gnu
                # style ld, but for clang on "real" windows we'll use
                # either link.exe or lld-link.exe
                try:
                    linker = guess_win_linker(env, compiler, cls, version, for_machine, invoked_directly=False)
                except MesonException:
                    pass
            if linker is None:
                linker = guess_nix_linker(env, compiler, cls, version, for_machine)

            return cls(
                ccache, compiler, version, for_machine, env,
                defines=defines, full_version=full_version, linker=linker)

        if 'Intel(R) C++ Intel(R)' in err:
            version = search_version(err)
            target = 'x86' if 'IA-32' in err else 'x86_64'
            cls = c.IntelClCCompiler if lang == 'c' else cpp.IntelClCPPCompiler
            env.add_lang_args(cls.language, cls, for_machine)
            linker = linkers.XilinkDynamicLinker(env, for_machine, [], version=version)
            return cls(
                compiler, version, for_machine, env, target,
                linker=linker)
        if 'Intel(R) oneAPI DPC++/C++ Compiler for applications' in err:
            version = search_version(err)
            target = 'x86' if 'IA-32' in err else 'x86_64'
            cls = c.IntelLLVMClCCompiler if lang == 'c' else cpp.IntelLLVMClCPPCompiler
            env.add_lang_args(cls.language, cls, for_machine)
            linker = linkers.XilinkDynamicLinker(env, for_machine, [], version=version)
            return cls(
                compiler, version, for_machine, env, target,
                linker=linker)
        if 'Microsoft' in out or 'Microsoft' in err:
            # Latest versions of Visual Studio print version
            # number to stderr but earlier ones print version
            # on stdout.  Why? Lord only knows.
            # Check both outputs to figure out version.
            for lookat in [err, out]:
                version = search_version(lookat)
                if version != 'unknown version':
                    break
            else:
                raise EnvironmentException(f'Failed to detect MSVC compiler version: stderr was\n{err!r}')
            cl_signature = lookat.split('\n', maxsplit=1)[0]
            match = re.search(r'.*(x86|x64|ARM|ARM64)([^_A-Za-z0-9]|$)', cl_signature)
            if match:
                target = match.group(1)
            else:
                m = f'Failed to detect MSVC compiler target architecture: \'cl /?\' output is\n{cl_signature}'
                raise EnvironmentException(m)
            cls = c.VisualStudioCCompiler if lang == 'c' else cpp.VisualStudioCPPCompiler
            linker = guess_win_linker(env, ['link'], cls, version, for_machine)
            if ccache_exe and ccache_exe.found():
                if ccache_exe.get_name() == 'ccache' and version_compare(ccache_exe.get_version(), '< 4.6'):
                    mlog.warning('Visual Studio support requires ccache 4.6 or higher. You have ccache {}. '.format(ccache_exe.get_version()), once=True)
                    ccache = []
            return cls(
                ccache, compiler, version, for_machine, env, target,
                full_version=cl_signature, linker=linker)
        if 'PGI Compilers' in out:
            cls = c.PGICCompiler if lang == 'c' else cpp.PGICPPCompiler
            env.add_lang_args(cls.language, cls, for_machine)
            linker = linkers.PGIDynamicLinker(compiler, env, for_machine, cls.LINKER_PREFIX, [], version=version)
            return cls(
                ccache, compiler, version, for_machine,
                env, linker=linker)
        if 'NVIDIA Compilers and Tools' in out:
            cls = c.NvidiaHPC_CCompiler if lang == 'c' else cpp.NvidiaHPC_CPPCompiler
            env.add_lang_args(cls.language, cls, for_machine)
            linker = linkers.NvidiaHPC_DynamicLinker(compiler, env, for_machine, cls.LINKER_PREFIX, [], version=version)
            return cls(
                ccache, compiler, version, for_machine,
                env, linker=linker)
        if '(ICC)' in out:
            cls = c.IntelCCompiler if lang == 'c' else cpp.IntelCPPCompiler
            l = guess_nix_linker(env, compiler, cls, version, for_machine)
            return cls(
                ccache, compiler, version, for_machine, env,
                full_version=full_version, linker=l)
        if 'Intel(R) oneAPI' in out:
            cls = c.IntelLLVMCCompiler if lang == 'c' else cpp.IntelLLVMCPPCompiler
            l = guess_nix_linker(env, compiler, cls, version, for_machine)
            return cls(
                ccache, compiler, version, for_machine, env,
                full_version=full_version, linker=l)
        if 'ARM' in out and not ('Metrowerks' in out or 'Freescale' in out):
            cls = c.ArmCCompiler if lang == 'c' else cpp.ArmCPPCompiler
            env.add_lang_args(cls.language, cls, for_machine)
            linker = linkers.ArmDynamicLinker(env, for_machine, version=version)
            return cls(
                ccache, compiler, version, for_machine,
                env, full_version=full_version, linker=linker)
        if 'RX Family' in out:
            cls = c.CcrxCCompiler if lang == 'c' else cpp.CcrxCPPCompiler
            env.add_lang_args(cls.language, cls, for_machine)
            linker = linkers.CcrxDynamicLinker(env, for_machine, version=version)
            return cls(
                ccache, compiler, version, for_machine, env,
                full_version=full_version, linker=linker)

        if 'Microchip' in out:
            if 'XC32' in out:
                # XC32 versions always have the form 'vMAJOR.MINOR'
                match = re.search(r'XC32.*v(\d+\.\d+)', out)
                if match:
                    version = match.group(1)
                else:
                    raise EnvironmentException(f'Failed to detect XC32 compiler version: full version was\n{full_version}')

                cls = c.Xc32CCompiler if lang == 'c' else cpp.Xc32CPPCompiler
                defines = _get_gnu_compiler_defines(compiler, lang)
                cls.gcc_version = _get_gnu_version_from_defines(defines)

                env.add_lang_args(cls.language, cls, for_machine)
                linker = linkers.Xc32DynamicLinker(compiler, env, for_machine, cls.LINKER_PREFIX, [], version=version)

                return cls(
                    ccache, compiler, version, for_machine,
                    env, defines=defines, full_version=full_version,
                    linker=linker)
            else:
                cls = c.Xc16CCompiler
                env.add_lang_args(cls.language, cls, for_machine)
                linker = linkers.Xc16DynamicLinker(env, for_machine, version=version)

                return cls(
                    ccache, compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

        if 'CompCert' in out:
            cls = c.CompCertCCompiler
            env.add_lang_args(cls.language, cls, for_machine)
            linker = linkers.CompCertDynamicLinker(env, for_machine, version=version)
            return cls(
                ccache, compiler, version, for_machine, env,
                full_version=full_version, linker=linker)

        if 'Metrowerks C/C++' in out or 'Freescale C/C++' in out:
            if 'ARM' in out:
                cls = c.MetrowerksCCompilerARM if lang == 'c' else cpp.MetrowerksCPPCompilerARM
                lnk = linkers.MetrowerksLinkerARM
            else:
                cls = c.MetrowerksCCompilerEmbeddedPowerPC if lang == 'c' else cpp.MetrowerksCPPCompilerEmbeddedPowerPC
                lnk = linkers.MetrowerksLinkerEmbeddedPowerPC

            mwcc_ver_match = re.search(r'Version (\d+)\.(\d+)\.?(\d+)? build (\d+)', out)
            assert mwcc_ver_match is not None, 'for mypy'  # because mypy *should* be complaining that this could be None
            compiler_version = '.'.join(x for x in mwcc_ver_match.groups() if x is not None)

            env.add_lang_args(cls.language, cls, for_machine)
            ld = env.lookup_binary_entry(for_machine, cls.language + '_ld')

            if ld is not None:
                _, o_ld, _ = Popen_safe(ld + ['--version'])

                mwld_ver_match = re.search(r'Version (\d+)\.(\d+)\.?(\d+)? build (\d+)', o_ld)
                assert mwld_ver_match is not None, 'for mypy'  # because mypy *should* be complaining that this could be None
                linker_version = '.'.join(x for x in mwld_ver_match.groups() if x is not None)

                linker = lnk(ld, env, for_machine, version=linker_version)
            else:
                raise EnvironmentException(f'Failed to detect linker for {cls.id!r} compiler. Please update your cross file(s).')

            return cls(
                ccache, compiler, compiler_version, for_machine, env,
                full_version=full_version, linker=linker)
        if 'TASKING VX-toolset' in err:
            cls = c.TaskingCCompiler
            lnk = linkers.TaskingLinker

            tasking_ver_match = re.search(r'v([0-9]+)\.([0-9]+)r([0-9]+) Build ([0-9]+)', err)
            assert tasking_ver_match is not None, 'for mypy'
            tasking_version = '.'.join(x for x in tasking_ver_match.groups() if x is not None)

            env.add_lang_args(cls.language, cls, for_machine)
            ld = env.lookup_binary_entry(for_machine, cls.language + '_ld')
            if ld is None:
                raise MesonException(f'{cls.language}_ld was not properly defined in your cross file')

            linker = lnk(ld, env, for_machine, version=tasking_version)
            return cls(
                ccache, compiler, tasking_version, for_machine, env,
                full_version=full_version, linker=linker)

    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException(f'Unknown compiler {compilers}')

def detect_c_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    return _detect_c_or_cpp_compiler(env, 'c', for_machine)

def detect_cpp_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    return _detect_c_or_cpp_compiler(env, 'cpp', for_machine)

def detect_cuda_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from .cuda import CudaCompiler, Phase
    from ..options import OptionKey
    from ..linkers.linkers import CudaLinker
    popen_exceptions = {}
    compilers, ccache_exe = _get_compilers(env, 'cuda', for_machine)
    ccache = ccache_exe.get_command() if (ccache_exe and ccache_exe.found()) else []
    for compiler in compilers:
        arg = '--version'
        try:
            p, out, err = Popen_safe_logged(compiler + [arg], msg='Detecting compiler via')
        except OSError as e:
            popen_exceptions[join_args(compiler + [arg])] = e
            continue
        # Example nvcc printout:
        #
        #     nvcc: NVIDIA (R) Cuda compiler driver
        #     Copyright (c) 2005-2018 NVIDIA Corporation
        #     Built on Sat_Aug_25_21:08:01_CDT_2018
        #     Cuda compilation tools, release 10.0, V10.0.130
        #
        # search_version() first finds the "10.0" after "release",
        # rather than the more precise "10.0.130" after "V".
        # The patch version number is occasionally important; For
        # instance, on Linux,
        #    - CUDA Toolkit 8.0.44 requires NVIDIA Driver 367.48
        #    - CUDA Toolkit 8.0.61 requires NVIDIA Driver 375.26
        # Luckily, the "V" also makes it very simple to extract
        # the full version:
        version = out.strip().rsplit('V', maxsplit=1)[-1]
        cpp_compiler = detect_cpp_compiler(env, for_machine)
        cls = CudaCompiler
        env.add_lang_args(cls.language, cls, for_machine)
        key = OptionKey('cuda_link_args', machine=for_machine)
        if not env.is_cross_build(for_machine):
            key = key.as_host()
        if key in env.options:
            # To fix LDFLAGS issue
            val = env.options[key]
            assert isinstance(val, list)
            env.coredata.optstore.set_option(key, cls.to_host_flags_base(val, Phase.LINKER))
        linker = CudaLinker(compiler, env, for_machine, CudaCompiler.LINKER_PREFIX, [], version=CudaLinker.parse_version())
        return cls(ccache, compiler, version, for_machine, cpp_compiler, env, linker=linker)
    raise EnvironmentException(f'Could not find suitable CUDA compiler: "{"; ".join([" ".join(c) for c in compilers])}"')

def detect_fortran_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from . import fortran
    from ..linkers import linkers
    popen_exceptions: T.Dict[str, T.Union[Exception, str]] = {}
    compilers, ccache = _get_compilers(env, 'fortran', for_machine)
    cls: T.Type[FortranCompiler]
    for compiler in compilers:
        # capture help text for possible fallback
        try:
            _, help_out, _ = Popen_safe_logged(compiler + ['--help'], msg='Detecting compiler via')
        except OSError as e:
            popen_exceptions[join_args(compiler + ['--help'])] = e
            help_out = ''

        for arg in ['--version', '-V']:
            try:
                p, out, err = Popen_safe_logged(compiler + [arg], msg='Detecting compiler via')
            except OSError as e:
                popen_exceptions[join_args(compiler + [arg])] = e
                continue

            version = search_version(out)
            full_version = out.split('\n', 1)[0]

            guess_gcc_or_lcc: T.Optional[str] = None
            if 'GNU Fortran' in out:
                guess_gcc_or_lcc = 'gcc'
            if 'e2k' in out and 'lcc' in out:
                guess_gcc_or_lcc = 'lcc'

            if guess_gcc_or_lcc:
                defines = _get_gnu_compiler_defines(compiler, 'fortran')
                if not defines:
                    popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                    continue
                if guess_gcc_or_lcc == 'lcc':
                    version = _get_lcc_version_from_defines(defines)
                    cls = fortran.ElbrusFortranCompiler
                    linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                    return cls(
                        compiler, version, for_machine, env,
                        defines, full_version=full_version, linker=linker)
                else:
                    version = _get_gnu_version_from_defines(defines)
                    cls = fortran.GnuFortranCompiler
                    linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                    return cls(
                        compiler, version, for_machine, env,
                        defines, full_version=full_version, linker=linker)

            if 'Arm C/C++/Fortran Compiler' in out:
                cls = fortran.ArmLtdFlangFortranCompiler
                arm_ver_match = re.search(r'version (\d+)\.(\d+)\.?(\d+)? \(build number (\d+)\)', out)
                assert arm_ver_match is not None, 'for mypy'  # because mypy *should* be complaining that this could be None
                version = '.'.join([x for x in arm_ver_match.groups() if x is not None])
                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                return cls(
                    compiler, version, for_machine, env,
                    linker=linker)
            if 'G95' in out:
                cls = fortran.G95FortranCompiler
                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'Sun Fortran' in err:
                version = search_version(err)
                cls = fortran.SunFortranCompiler
                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'Intel(R) Fortran Compiler for applications' in err:
                version = search_version(err)
                target = 'x86' if 'IA-32' in err else 'x86_64'
                cls = fortran.IntelLLVMClFortranCompiler
                env.add_lang_args(cls.language, cls, for_machine)
                linker = linkers.XilinkDynamicLinker(env, for_machine, [], version=version)
                return cls(
                    compiler, version, for_machine, env,
                    target, linker=linker)

            if 'Intel(R) Visual Fortran' in err or 'Intel(R) Fortran' in err:
                version = search_version(err)
                target = 'x86' if 'IA-32' in err else 'x86_64'
                cls = fortran.IntelClFortranCompiler
                env.add_lang_args(cls.language, cls, for_machine)
                linker = linkers.XilinkDynamicLinker(env, for_machine, [], version=version)
                return cls(
                    compiler, version, for_machine, env,
                    target, linker=linker)

            if 'ifort (IFORT)' in out:
                cls = fortran.IntelFortranCompiler
                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'ifx (IFORT)' in out or 'ifx (IFX)' in out:
                cls = fortran.IntelLLVMFortranCompiler
                linker = guess_nix_linker(env, compiler, cls, version, for_machine)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'PathScale EKOPath(tm)' in err:
                return fortran.PathScaleFortranCompiler(
                    compiler, version, for_machine, env,
                    full_version=full_version)

            if 'PGI Compilers' in out:
                cls = fortran.PGIFortranCompiler
                env.add_lang_args(cls.language, cls, for_machine)
                linker = linkers.PGIDynamicLinker(compiler, env, for_machine,
                                                  cls.LINKER_PREFIX, [], version=version)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'NVIDIA Compilers and Tools' in out:
                cls = fortran.NvidiaHPC_FortranCompiler
                env.add_lang_args(cls.language, cls, for_machine)
                linker = linkers.PGIDynamicLinker(compiler, env, for_machine,
                                                  cls.LINKER_PREFIX, [], version=version)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            def _get_linker_try_windows(cls: T.Type['Compiler']) -> T.Optional['DynamicLinker']:
                linker = None
                if 'windows' in out or env.machines[for_machine].is_windows():
                    # If we're in a MINGW context this actually will use a gnu
                    # style ld, but for flang on "real" windows we'll use
                    # either link.exe or lld-link.exe
                    try:
                        linker = guess_win_linker(
                            env, compiler, cls, version,
                            for_machine, invoked_directly=False
                        )
                    except MesonException:
                        pass
                if linker is None:
                    linker = guess_nix_linker(env, compiler, cls,
                                              version, for_machine)
                return linker

            if 'flang-new' in out or 'flang LLVM compiler' in help_out:
                cls = fortran.LlvmFlangFortranCompiler
                linker = _get_linker_try_windows(cls)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'flang' in out or 'clang' in out:
                cls = fortran.ClassicFlangFortranCompiler
                linker = _get_linker_try_windows(cls)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'Open64 Compiler Suite' in err:
                cls = fortran.Open64FortranCompiler
                linker = guess_nix_linker(env,
                                          compiler, cls, version, for_machine)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

            if 'NAG Fortran' in err:
                full_version = err.split('\n', 1)[0]
                version = full_version.split()[-1]
                cls = fortran.NAGFortranCompiler
                env.add_lang_args(cls.language, cls, for_machine)
                linker = linkers.NAGDynamicLinker(
                    compiler, env, for_machine, cls.LINKER_PREFIX, [],
                    version=version)
                return cls(
                    compiler, version, for_machine, env,
                    full_version=full_version, linker=linker)

    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_objc_compiler(env: 'Environment', for_machine: MachineChoice) -> 'Compiler':
    return _detect_objc_or_objcpp_compiler(env, 'objc', for_machine)

def detect_objcpp_compiler(env: 'Environment', for_machine: MachineChoice) -> 'Compiler':
    return _detect_objc_or_objcpp_compiler(env, 'objcpp', for_machine)

def _detect_objc_or_objcpp_compiler(env: 'Environment', lang: str, for_machine: MachineChoice) -> 'Compiler':
    from . import objc, objcpp
    popen_exceptions: T.Dict[str, T.Union[Exception, str]] = {}
    compilers, ccache_exe = _get_compilers(env, lang, for_machine)
    ccache = ccache_exe.get_command() if (ccache_exe and ccache_exe.found()) else []
    comp: T.Union[T.Type[objc.ObjCCompiler], T.Type[objcpp.ObjCPPCompiler]]

    for compiler in compilers:
        arg = ['--version']
        try:
            p, out, err = Popen_safe_logged(compiler + arg, msg='Detecting compiler via')
        except OSError as e:
            popen_exceptions[join_args(compiler + arg)] = e
            continue
        version = search_version(out)
        if 'Free Software Foundation' in out:
            defines = _get_gnu_compiler_defines(compiler, lang)
            if not defines:
                popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                continue
            version = _get_gnu_version_from_defines(defines)
            comp = objc.GnuObjCCompiler if lang == 'objc' else objcpp.GnuObjCPPCompiler
            linker = guess_nix_linker(env, compiler, comp, version, for_machine)
            c = comp(
                ccache, compiler, version, for_machine, env,
                defines, linker=linker)
            if not c.compiles('int main(void) { return 0; }')[0]:
                popen_exceptions[join_args(compiler)] = f'GCC was not built with support for {"objective-c" if lang == "objc" else "objective-c++"}'
                continue
            return c
        if 'clang' in out:
            linker = None
            defines = _get_clang_compiler_defines(compiler, lang)
            if not defines:
                popen_exceptions[join_args(compiler)] = 'no pre-processor defines'
                continue
            if 'Apple' in out:
                comp = objc.AppleClangObjCCompiler if lang == 'objc' else objcpp.AppleClangObjCPPCompiler
            else:
                comp = objc.ClangObjCCompiler if lang == 'objc' else objcpp.ClangObjCPPCompiler
            if 'windows' in out or env.machines[for_machine].is_windows():
                # If we're in a MINGW context this actually will use a gnu style ld
                try:
                    linker = guess_win_linker(env, compiler, comp, version, for_machine)
                except MesonException:
                    pass

            if not linker:
                linker = guess_nix_linker(env, compiler, comp, version, for_machine)
            return comp(
                ccache, compiler, version, for_machine,
                env, linker=linker, defines=defines)
    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_java_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from .java import JavaCompiler
    exelist = env.lookup_binary_entry(for_machine, 'java')
    if exelist is None:
        # TODO support fallback
        exelist = [defaults['java'][0]]

    try:
        p, out, err = Popen_safe_logged(exelist + ['-version'], msg='Detecting compiler via')
    except OSError:
        raise EnvironmentException('Could not execute Java compiler: {}'.format(join_args(exelist)))
    if 'javac' in out or 'javac' in err:
        version = search_version(err if 'javac' in err else out)
        if not version or version == 'unknown version':
            parts = (err if 'javac' in err else out).split()
            if len(parts) > 1:
                version = parts[1]
        comp_class = JavaCompiler
        env.add_lang_args(comp_class.language, comp_class, for_machine)
        return comp_class(exelist, version, for_machine, env)
    raise EnvironmentException('Unknown compiler: ' + join_args(exelist))

def detect_cs_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from . import cs
    compilers, ccache = _get_compilers(env, 'cs', for_machine)
    popen_exceptions = {}
    for comp in compilers:
        try:
            p, out, err = Popen_safe_logged(comp + ['--version'], msg='Detecting compiler via')
        except OSError as e:
            popen_exceptions[join_args(comp + ['--version'])] = e
            continue

        version = search_version(out)
        cls: T.Type[cs.CsCompiler]
        if 'Mono' in out:
            cls = cs.MonoCompiler
        elif "Visual C#" in out:
            cls = cs.VisualStudioCsCompiler
        else:
            continue
        env.add_lang_args(cls.language, cls, for_machine)
        return cls(comp, version, for_machine, env)

    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_cython_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    """Search for a cython compiler."""
    from .cython import CythonCompiler
    compilers, _ = _get_compilers(env, 'cython', MachineChoice.BUILD)

    popen_exceptions: T.Dict[str, Exception] = {}
    for comp in compilers:
        try:
            _, out, err = Popen_safe_logged(comp + ['-V'], msg='Detecting compiler via')
        except OSError as e:
            popen_exceptions[join_args(comp + ['-V'])] = e
            continue

        version: T.Optional[str] = None
        # 3.0
        if 'Cython' in out:
            version = search_version(out)
        # older
        elif 'Cython' in err:
            version = search_version(err)
        if version is not None:
            comp_class = CythonCompiler
            env.add_lang_args(comp_class.language, comp_class, for_machine)
            return comp_class([], comp, version, for_machine, env)
    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_vala_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from .vala import ValaCompiler
    exelist = env.lookup_binary_entry(MachineChoice.BUILD, 'vala')
    if exelist is None:
        # TODO support fallback
        exelist = [defaults['vala'][0]]

    try:
        p, out = Popen_safe_logged(exelist + ['--version'], msg='Detecting compiler via')[0:2]
    except OSError:
        raise EnvironmentException('Could not execute Vala compiler: {}'.format(join_args(exelist)))
    version = search_version(out)
    if 'Vala' in out:
        comp_class = ValaCompiler
        env.add_lang_args(comp_class.language, comp_class, for_machine)
        return comp_class(exelist, version, for_machine, env)
    raise EnvironmentException('Unknown compiler: ' + join_args(exelist))

def detect_rust_compiler(env: 'Environment', for_machine: MachineChoice) -> RustCompiler:
    from . import rust
    from ..linkers import linkers
    popen_exceptions: T.Dict[str, Exception] = {}
    compilers, _ = _get_compilers(env, 'rust', for_machine)
    override = env.lookup_binary_entry(for_machine, 'rust_ld')

    for compiler in compilers:
        arg = ['--version']
        try:
            out = Popen_safe_logged(compiler + arg, msg='Detecting compiler via')[1]
        except OSError as e:
            popen_exceptions[join_args(compiler + arg)] = e
            continue

        # Full version contains the "-nightly" or "-beta" suffixes, but version
        # should just be X.Y.Z
        full_version = search_version(out)
        version = full_version.split('-', 1)[0]

        cls: T.Type[RustCompiler] = rust.RustCompiler

        # Clippy is a wrapper around rustc, but it doesn't have rustc in its
        # output. We can otherwise treat it as rustc.
        if 'clippy' in out:
            # clippy returns its own version and not the rustc version by
            # default so try harder here to get the correct version.
            # Also replace the whole output with the rustc output in
            # case this is later used for other purposes.
            arg = ['--rustc', '--version']
            try:
                out = Popen_safe(compiler + arg)[1]
            except OSError as e:
                popen_exceptions[join_args(compiler + arg)] = e
                continue
            full_version = search_version(out)
            version = full_version.split('-', 1)[0]

            cls = rust.ClippyRustCompiler
            mlog.deprecation(
                'clippy-driver is not intended as a general purpose compiler. '
                'You can use "ninja clippy" in order to run clippy on a '
                'meson project.')

        if 'rustc' in out:
            # On Linux and mac rustc will invoke gcc (clang for mac
            # presumably) and it can do this windows, for dynamic linking.
            # this means the easiest way to figure out what linker to use
            # is to just get the value of the C compiler and use that as
            # the basis of the rust linker.
            #
            # However, there are two things we need to change, if CC is not
            # the default use that, and second add the necessary arguments
            # to rust to use -fuse-ld
            #
            # For MSVC targets, require an MSVC-compatible C compiler to get
            # the corresponding linker
            rust_target = rust.parse_target(compiler)
            if rust_target and rust_target.endswith('-msvc'):
                try:
                    cc = _detect_c_or_cpp_compiler(env, 'c', for_machine,
                                                   override_compiler=['clang-cl', 'cl'])
                except EnvironmentException:
                    popen_exceptions[join_args(compiler)] = \
                        EnvironmentException('No MSVC-compatible C compiler found for MSVC Rust target')
                    continue
            else:
                try:
                    cc = detect_c_compiler(env, for_machine)
                except EnvironmentException as e:
                    popen_exceptions[join_args(compiler)] = e
                    continue

            is_link_exe = isinstance(cc.linker, linkers.VisualStudioLikeLinkerMixin)

            if any(a.startswith('linker=') for a in compiler):
                mlog.warning(
                    'Please do not put -C linker= in your compiler '
                    'command, set rust_ld=command in your cross file '
                    'or use the RUSTC_LD environment variable, otherwise meson '
                    'will override your selection.')

            compiler = compiler.copy()  # avoid mutating the original list

            if override is None:
                assert cc.linker is not None, 'for mypy'
                linker: DynamicLinker
                exelist = cc.get_linker_exelist()
                if is_link_exe:
                    # This trickery with type() gets us the class of the linker
                    # so we can initialize a new copy for the Rust Compiler
                    # Due to initializer mismatch we can't use the VisualStudioLikeMixin here
                    # But all of these have the same API so we can just pick one.
                    linker = T.cast('T.Type[linkers.MSVCDynamicLinker]', type(cc.linker))(
                        env, for_machine, always_args=[],
                        exelist=exelist, version=cc.linker.version,
                        direct=True, machine=cc.linker.machine)
                else:
                    linker = type(cc.linker)(exelist, env, for_machine, cc.LINKER_PREFIX,
                                             always_args=[], system=cc.linker.system,
                                             version=cc.linker.version)
                    exelist += cc.linker.get_always_args()
            elif 'link' in override[0]:
                linker = guess_win_linker(env,
                                          override, cls, version, for_machine, use_linker_prefix=False)
                # rustc takes linker arguments without a prefix, and
                # inserts the correct prefix itself.
                assert isinstance(linker, linkers.VisualStudioLikeLinkerMixin)
                linker.direct = True
                exelist = linker.get_exelist()
            else:
                # On linux and macos rust will invoke the c compiler for
                # linking, on windows it will use lld-link or link.exe.
                # we will simply ask for the C compiler that corresponds to
                # it, and use that.
                cc = _detect_c_or_cpp_compiler(env, 'c', for_machine, override_compiler=override)
                linker = cc.linker
                exelist = cc.get_linker_exelist()

            c = exelist.pop(0)
            compiler.extend(cls.use_linker_args(c, ''))
            compiler.extend(rust.rustc_link_args(exelist))

            env.add_lang_args(cls.language, cls, for_machine)
            return cls(
                compiler, version, for_machine, env,
                linker=linker, full_version=full_version)

    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_d_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from . import c, d

    # Detect the target architecture, required for proper architecture handling on Windows.
    # MSVC compiler is required for correct platform detection.
    c_compiler = {'c': detect_c_compiler(env, for_machine)}
    is_msvc = isinstance(c_compiler['c'], c.VisualStudioCCompiler)
    if not is_msvc:
        c_compiler = {}

    arch = detect_cpu_family(c_compiler)
    if is_msvc and arch == 'x86':
        arch = 'x86_mscoff'

    popen_exceptions = {}
    info = env.machines[for_machine]
    compilers, ccache = _get_compilers(env, 'd', for_machine)
    cls: T.Type[d.DCompiler]
    for exelist in compilers:
        # Search for a D compiler.
        # We prefer LDC over GDC unless overridden with the DC
        # environment variable because LDC has a much more
        # up to date language version at time (2016).
        if os.path.basename(exelist[-1]).startswith(('ldmd', 'gdmd')):
            raise EnvironmentException(
                f'Meson does not support {exelist[-1]} as it is only a DMD frontend for another compiler.'
                'Please provide a valid value for DC or unset it so that Meson can resolve the compiler by itself.')
        try:
            p, out = Popen_safe(exelist + ['--version'])[0:2]
        except OSError as e:
            popen_exceptions[join_args(exelist + ['--version'])] = e
            continue
        version = search_version(out)
        full_version = out.split('\n', 1)[0]

        # The OpenD fork should stay close enough to upstream D (in
        # the areas that interest us) to allow supporting them both
        # without much hassle.
        # See: https://github.com/orgs/opendlang/discussions/56
        if 'LLVM D compiler' in out or 'LLVM Open D compiler' in out:
            cls = d.LLVMDCompiler
            # LDC seems to require a file
            # We cannot use NamedTemporaryFile on windows, its documented
            # to not work for our uses. So, just use mkstemp and only have
            # one path for simplicity.
            o, f = tempfile.mkstemp('.d')
            os.close(o)

            try:
                if info.is_windows() or info.is_cygwin():
                    objfile = os.path.basename(f)[:-1] + 'obj'
                    extra_args = [f]
                    if env.is_cross_build(for_machine):
                        extra_args.append(f'-mtriple={info.cpu}-windows')

                    linker = guess_win_linker(env,
                                              exelist,
                                              cls, full_version, for_machine,
                                              use_linker_prefix=True, invoked_directly=False,
                                              extra_args=extra_args)
                else:
                    # LDC writes an object file to the current working directory.
                    # Clean it up.
                    objfile = os.path.basename(f)[:-1] + 'o'
                    linker = guess_nix_linker(env,
                                              exelist, cls, full_version, for_machine,
                                              extra_args=[f])
            finally:
                windows_proof_rm(f)
                windows_proof_rm(objfile)

            return cls(
                exelist, version, for_machine, env, arch,
                full_version=full_version, linker=linker,
                version_output=out)
        elif 'gdc' in out:
            cls = d.GnuDCompiler
            linker = guess_nix_linker(env, exelist, cls, version, for_machine)
            return cls(
                exelist, version, for_machine, env, arch,
                full_version=full_version, linker=linker)
        elif 'The D Language Foundation' in out or 'Digital Mars' in out:
            cls = d.DmdDCompiler
            # DMD seems to require a file
            # We cannot use NamedTemporaryFile on windows, its documented
            # to not work for our uses. So, just use mkstemp and only have
            # one path for simplicity.
            o, f = tempfile.mkstemp('.d')
            os.close(o)

            # DMD as different detection logic for x86 and x86_64
            arch_arg = '-m64' if arch == 'x86_64' else '-m32'

            try:
                if info.is_windows() or info.is_cygwin():
                    objfile = os.path.basename(f)[:-1] + 'obj'
                    linker = guess_win_linker(env,
                                              exelist, cls, full_version, for_machine,
                                              invoked_directly=False, extra_args=[f, arch_arg])
                else:
                    objfile = os.path.basename(f)[:-1] + 'o'
                    linker = guess_nix_linker(env,
                                              exelist, cls, full_version, for_machine,
                                              extra_args=[f, arch_arg])
            finally:
                windows_proof_rm(f)
                windows_proof_rm(objfile)

            return cls(
                exelist, version, for_machine, env, arch,
                full_version=full_version, linker=linker)
        raise EnvironmentException('Unknown compiler: ' + join_args(exelist))

    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_swift_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from .swift import SwiftCompiler
    exelist = env.lookup_binary_entry(for_machine, 'swift')
    if exelist is None:
        # TODO support fallback
        exelist = [defaults['swift'][0]]

    try:
        p, _, err = Popen_safe_logged(exelist + ['-v'], msg='Detecting compiler via')
    except OSError:
        raise EnvironmentException('Could not execute Swift compiler: {}'.format(join_args(exelist)))
    version = search_version(err)
    if 'Swift' in err:
        # As for 5.0.1 swiftc *requires* a file to check the linker:
        with tempfile.NamedTemporaryFile(suffix='.swift') as f:
            cls = SwiftCompiler
            linker = guess_nix_linker(env,
                                      exelist, cls, version, for_machine,
                                      extra_args=[f.name, '-o', '/dev/null'])
        return cls(
            exelist, version, for_machine, env, linker=linker)

    raise EnvironmentException('Unknown compiler: ' + join_args(exelist))

def detect_nasm_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    from .asm import NasmCompiler, YasmCompiler, MetrowerksAsmCompilerARM, MetrowerksAsmCompilerEmbeddedPowerPC

    # When cross compiling and nasm is not defined in the cross file we can
    # fallback to the build machine nasm.
    compilers, _ = _get_compilers(env, 'nasm', for_machine, allow_build_machine=True)

    # We need a C compiler to properly detect the machine info and linker
    cc = detect_c_compiler(env, for_machine)
    if not env.is_cross_build(for_machine):
        env.update_build_machine({'c': cc})

    popen_exceptions: T.Dict[str, Exception] = {}
    for comp in compilers:
        if comp == ['nasm'] and is_windows() and not shutil.which(comp[0]):
            # nasm is not in PATH on Windows by default
            default_path = os.path.join(os.environ['ProgramFiles'], 'NASM')
            comp[0] = shutil.which(comp[0], path=default_path) or comp[0]
        try:
            output = Popen_safe_logged(comp + ['--version'], msg='Detecting compiler via')[1]
        except OSError as e:
            popen_exceptions[' '.join(comp + ['--version'])] = e
            continue

        version = search_version(output)
        comp_class: T.Type[ASMCompiler]
        if 'NASM' in output:
            comp_class = NasmCompiler
            env.add_lang_args(comp_class.language, comp_class, for_machine)
            return comp_class([], comp, version, for_machine, env, cc.linker)
        elif 'yasm' in output:
            comp_class = YasmCompiler
            env.add_lang_args(comp_class.language, comp_class, for_machine)
            return comp_class([], comp, version, for_machine, env, cc.linker)
        elif 'Metrowerks' in output or 'Freescale' in output:
            if 'ARM' in output:
                comp_class_mwasmarm = MetrowerksAsmCompilerARM
                env.add_lang_args(comp_class_mwasmarm.language, comp_class_mwasmarm, for_machine)
                return comp_class_mwasmarm([], comp, version, for_machine, env, cc.linker)
            else:
                comp_class_mwasmeppc = MetrowerksAsmCompilerEmbeddedPowerPC
                env.add_lang_args(comp_class_mwasmeppc.language, comp_class_mwasmeppc, for_machine)
                return comp_class_mwasmeppc([], comp, version, for_machine, env, cc.linker)

    _handle_exceptions(popen_exceptions, compilers)
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_masm_compiler(env: 'Environment', for_machine: MachineChoice) -> Compiler:
    # We need a C compiler to properly detect the machine info and linker
    cc = detect_c_compiler(env, for_machine)
    if not env.is_cross_build(for_machine):
        env.update_build_machine({'c': cc})

    info = env.machines[for_machine]

    from .asm import MasmCompiler, MasmARMCompiler
    comp_class: T.Type[ASMCompiler]
    if info.cpu_family == 'x86':
        comp = ['ml']
        comp_class = MasmCompiler
        arg = '/?'
    elif info.cpu_family == 'x86_64':
        comp = ['ml64']
        comp_class = MasmCompiler
        arg = '/?'
    elif info.cpu_family == 'arm':
        comp = ['armasm']
        comp_class = MasmARMCompiler
        arg = '-h'
    elif info.cpu_family == 'aarch64':
        comp = ['armasm64']
        comp_class = MasmARMCompiler
        arg = '-h'
    else:
        raise EnvironmentException(f'Platform {info.cpu_family} not supported by MASM')

    popen_exceptions: T.Dict[str, Exception] = {}
    try:
        output = Popen_safe(comp + [arg])[2]
        version = search_version(output)
        env.add_lang_args(comp_class.language, comp_class, for_machine)
        return comp_class([], comp, version, for_machine, env, cc.linker)
    except OSError as e:
        popen_exceptions[' '.join(comp + [arg])] = e
    _handle_exceptions(popen_exceptions, [comp])
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

def detect_linearasm_compiler(env: Environment, for_machine: MachineChoice) -> Compiler:
    from .asm import TILinearAsmCompiler
    comp = ['cl6x']
    comp_class: T.Type[ASMCompiler] = TILinearAsmCompiler
    arg = '-h'
    cc = detect_c_compiler(env, for_machine)
    if not env.is_cross_build(for_machine):
        env.update_build_machine({'c': cc})

    popen_exceptions: T.Dict[str, Exception] = {}
    try:
        output = Popen_safe(comp + [arg])[2]
        version = search_version(output)
        env.add_lang_args(comp_class.language, comp_class, for_machine)
        return comp_class([], comp, version, for_machine, env, cc.linker)
    except OSError as e:
        popen_exceptions[' '.join(comp + [arg])] = e
    _handle_exceptions(popen_exceptions, [comp])
    raise EnvironmentException('Unreachable code (exception to make mypy happy)')

# GNU/Clang defines and version
# =============================

def _get_gnu_compiler_defines(compiler: T.List[str], lang: str) -> T.Dict[str, str]:
    """
    Get the list of GCC pre-processor defines
    """
    from .mixins.gnu import gnu_lang_map

    def _try_obtain_compiler_defines(args: T.List[str]) -> str:
        mlog.debug(f'Running command: {join_args(args)}')
        p, output, error = Popen_safe(compiler + args, write='', stdin=subprocess.PIPE)
        if p.returncode != 0:
            raise EnvironmentException('Unable to get gcc pre-processor defines:\n'
                                       f'Compiler stdout:\n{output}\n-----\n'
                                       f'Compiler stderr:\n{error}\n-----\n')
        return output

    # Arguments to output compiler pre-processor defines to stdout
    # gcc, g++, and gfortran all support these arguments
    baseline_test_args = ['-E', '-dM', '-']
    try:
        # We assume that when _get_gnu_compiler_defines is called, it's
        # close enough to a GCCish compiler so we reuse the _LANG_MAP
        # from the GCC mixin. This isn't a dangerous assumption because
        # we fallback if the detection fails anyway.

        # We might not have a match for Fortran, so fallback to detection
        # based on the driver.
        lang = gnu_lang_map[lang]

        # The compiler may not infer the target language based on the driver name
        # so first, try with '-cpp -x lang', then fallback without given it's less
        # portable. We try with '-cpp' as GCC needs it for Fortran at least, and
        # it seems to do no harm.
        output = _try_obtain_compiler_defines(['-cpp', '-x', lang] + baseline_test_args)
    except (EnvironmentException, KeyError):
        mlog.debug(f'pre-processor extraction using -cpp -x {lang} failed, falling back w/o lang')
        output = _try_obtain_compiler_defines(baseline_test_args)

    # Parse several lines of the type:
    # `#define ___SOME_DEF some_value`
    # and extract `___SOME_DEF`
    defines: T.Dict[str, str] = {}
    for line in output.split('\n'):
        if not line:
            continue
        d, *rest = line.split(' ', 2)
        if d != '#define':
            continue
        if len(rest) == 1:
            defines[rest[0]] = ''
        if len(rest) == 2:
            defines[rest[0]] = rest[1]
    return defines

def _get_clang_compiler_defines(compiler: T.List[str], lang: str) -> T.Dict[str, str]:
    """
    Get the list of Clang pre-processor defines
    """
    from .mixins.clang import clang_lang_map

    def _try_obtain_compiler_defines(args: T.List[str]) -> str:
        mlog.debug(f'Running command: {join_args(args)}')
        p, output, error = Popen_safe(compiler + args, write='', stdin=subprocess.PIPE)
        if p.returncode != 0:
            raise EnvironmentException('Unable to get clang pre-processor defines:\n'
                                       f'Compiler stdout:\n{output}\n-----\n'
                                       f'Compiler stderr:\n{error}\n-----\n')
        return output

    # Arguments to output compiler pre-processor defines to stdout
    baseline_test_args = ['-E', '-dM', '-']
    try:
        # We assume that when _get_clang_compiler_defines is called, it's
        # close enough to a Clangish compiler so we reuse the _LANG_MAP
        # from the Clang mixin. This isn't a dangerous assumption because
        # we fallback if the detection fails anyway.

        # We might not have a match for Fortran, so fallback to detection
        # based on the driver.
        lang = clang_lang_map[lang]

        # The compiler may not infer the target language based on the driver name.
        # Try first with '-x lang' to supported systemwide language level overrides,
        # then fallback to without since it's a more recent option.
        output = _try_obtain_compiler_defines(['-x', lang] + baseline_test_args)
    except (EnvironmentException, KeyError):
        mlog.debug(f'pre-processor extraction using -x {lang} failed, falling back w/o lang')
        output = _try_obtain_compiler_defines(baseline_test_args)

    defines: T.Dict[str, str] = {}
    for line in output.split('\n'):
        if not line:
            continue
        d, *rest = line.split(' ', 2)
        if d != '#define':
            continue
        if len(rest) == 1:
            defines[rest[0]] = ''
        if len(rest) == 2:
            defines[rest[0]] = rest[1]
    return defines

def _get_gnu_version_from_defines(defines: T.Dict[str, str]) -> str:
    dot = '.'
    major = defines.get('__GNUC__', '0')
    minor = defines.get('__GNUC_MINOR__', '0')
    patch = defines.get('__GNUC_PATCHLEVEL__', '0')
    return dot.join((major, minor, patch))

def _get_lcc_version_from_defines(defines: T.Dict[str, str]) -> str:
    dot = '.'
    generation_and_major = defines.get('__LCC__', '100')
    generation = generation_and_major[:1]
    major = generation_and_major[1:]
    minor = defines.get('__LCC_MINOR__', '0')
    return dot.join((generation, major, minor))



from __future__ import annotations

import os
from contextlib import contextmanager
from functools import cached_property
from typing import TYPE_CHECKING, Any, TypeVar

import pathspec

from hatchling.builders.constants import DEFAULT_BUILD_DIRECTORY, EXCLUDED_DIRECTORIES, BuildEnvVars
from hatchling.builders.utils import normalize_inclusion_map, normalize_relative_directory, normalize_relative_path
from hatchling.metadata.utils import normalize_project_name
from hatchling.utils.fs import locate_file

if TYPE_CHECKING:
    from collections.abc import Generator

    from hatchling.builders.plugin.interface import BuilderInterface


class BuilderConfig:
    def __init__(
        self,
        builder: BuilderInterface,
        root: str,
        plugin_name: str,
        build_config: dict[str, Any],
        target_config: dict[str, Any],
    ) -> None:
        self.__builder = builder
        self.__root = root
        self.__plugin_name = plugin_name
        self.__build_config = build_config
        self.__target_config = target_config

        # This is used when the only file selection is based on forced inclusion or build-time artifacts. This
        # instructs to `exclude` every encountered path without doing pattern matching that matches everything.
        self.__exclude_all: bool = False

        # Modified at build time
        self.build_artifact_spec: pathspec.GitIgnoreSpec | None = None
        self.build_force_include: dict[str, str] = {}
        self.build_reserved_paths: set[str] = set()

    @property
    def builder(self) -> BuilderInterface:
        return self.__builder

    @property
    def root(self) -> str:
        return self.__root

    @property
    def plugin_name(self) -> str:
        return self.__plugin_name

    @property
    def build_config(self) -> dict[str, Any]:
        return self.__build_config

    @property
    def target_config(self) -> dict[str, Any]:
        return self.__target_config

    def include_path(self, relative_path: str, *, explicit: bool = False, is_package: bool = True) -> bool:
        return (
            self.path_is_build_artifact(relative_path)
            or self.path_is_artifact(relative_path)
            or (
                not (self.only_packages and not is_package)
                and not self.path_is_excluded(relative_path)
                and (explicit or self.path_is_included(relative_path))
            )
        )

    def path_is_included(self, relative_path: str) -> bool:
        if self.include_spec is None:
            return True

        return self.include_spec.match_file(relative_path)

    def path_is_excluded(self, relative_path: str) -> bool:
        if self.__exclude_all:
            return True

        if self.exclude_spec is None:
            return False

        return self.exclude_spec.match_file(relative_path)

    def path_is_artifact(self, relative_path: str) -> bool:
        if self.artifact_spec is None:
            return False

        return self.artifact_spec.match_file(relative_path)

    def path_is_build_artifact(self, relative_path: str) -> bool:
        if self.build_artifact_spec is None:
            return False

        return self.build_artifact_spec.match_file(relative_path)

    def path_is_reserved(self, relative_path: str) -> bool:
        return relative_path in self.build_reserved_paths

    def directory_is_excluded(self, name: str, relative_path: str) -> bool:
        if name in EXCLUDED_DIRECTORIES:
            return True

        relative_directory = os.path.join(relative_path, name)
        return (
            self.path_is_reserved(relative_directory)
            # The trailing slash is necessary so e.g. `bar/` matches `foo/bar`
            or (self.skip_excluded_dirs and self.path_is_excluded(f"{relative_directory}/"))
        )

    @cached_property
    def include_spec(self) -> pathspec.GitIgnoreSpec | None:
        if "include" in self.target_config:
            include_config = self.target_config
            include_location = f"tool.hatch.build.targets.{self.plugin_name}.include"
        else:
            include_config = self.build_config
            include_location = "tool.hatch.build.include"

        all_include_patterns = []

        include_patterns = include_config.get("include", self.default_include())
        if not isinstance(include_patterns, list):
            message = f"Field `{include_location}` must be an array of strings"
            raise TypeError(message)

        for i, include_pattern in enumerate(include_patterns, 1):
            if not isinstance(include_pattern, str):
                message = f"Pattern #{i} in field `{include_location}` must be a string"
                raise TypeError(message)

            if not include_pattern:
                message = f"Pattern #{i} in field `{include_location}` cannot be an empty string"
                raise ValueError(message)

            all_include_patterns.append(include_pattern)

        # Matching only at the root requires a forward slash, back slashes do not work. As such,
        # normalize to forward slashes for consistency.
        all_include_patterns.extend(f"/{relative_path.replace(os.sep, '/')}/" for relative_path in self.packages)

        if all_include_patterns:
            return pathspec.GitIgnoreSpec.from_lines(all_include_patterns)
        return None

    @cached_property
    def exclude_spec(self) -> pathspec.GitIgnoreSpec | None:
        if "exclude" in self.target_config:
            exclude_config = self.target_config
            exclude_location = f"tool.hatch.build.targets.{self.plugin_name}.exclude"
        else:
            exclude_config = self.build_config
            exclude_location = "tool.hatch.build.exclude"

        all_exclude_patterns = self.default_global_exclude()

        if not self.ignore_vcs:
            all_exclude_patterns.extend(self.load_vcs_exclusion_patterns())

        exclude_patterns = exclude_config.get("exclude", self.default_exclude())
        if not isinstance(exclude_patterns, list):
            message = f"Field `{exclude_location}` must be an array of strings"
            raise TypeError(message)

        for i, exclude_pattern in enumerate(exclude_patterns, 1):
            if not isinstance(exclude_pattern, str):
                message = f"Pattern #{i} in field `{exclude_location}` must be a string"
                raise TypeError(message)

            if not exclude_pattern:
                message = f"Pattern #{i} in field `{exclude_location}` cannot be an empty string"
                raise ValueError(message)

            all_exclude_patterns.append(exclude_pattern)

        if all_exclude_patterns:
            return pathspec.GitIgnoreSpec.from_lines(all_exclude_patterns)
        return None

    @property
    def artifact_spec(self) -> pathspec.GitIgnoreSpec | None:
        if "artifacts" in self.target_config:
            artifact_config = self.target_config
            artifact_location = f"tool.hatch.build.targets.{self.plugin_name}.artifacts"
        else:
            artifact_config = self.build_config
            artifact_location = "tool.hatch.build.artifacts"

        all_artifact_patterns = []

        artifact_patterns = artifact_config.get("artifacts", [])
        if not isinstance(artifact_patterns, list):
            message = f"Field `{artifact_location}` must be an array of strings"
            raise TypeError(message)

        for i, artifact_pattern in enumerate(artifact_patterns, 1):
            if not isinstance(artifact_pattern, str):
                message = f"Pattern #{i} in field `{artifact_location}` must be a string"
                raise TypeError(message)

            if not artifact_pattern:
                message = f"Pattern #{i} in field `{artifact_location}` cannot be an empty string"
                raise ValueError(message)

            all_artifact_patterns.append(artifact_pattern)

        if all_artifact_patterns:
            return pathspec.GitIgnoreSpec.from_lines(all_artifact_patterns)
        return None

    @cached_property
    def hook_config(self) -> dict[str, Any]:
        hook_config: dict[str, dict[str, Any]] = {}

        global_hook_config = self.build_config.get("hooks", {})
        if not isinstance(global_hook_config, dict):
            message = "Field `tool.hatch.build.hooks` must be a table"
            raise TypeError(message)

        for hook_name, config in global_hook_config.items():
            if not isinstance(config, dict):
                message = f"Field `tool.hatch.build.hooks.{hook_name}` must be a table"
                raise TypeError(message)

            hook_config.setdefault(hook_name, config)

        target_hook_config = self.target_config.get("hooks", {})
        if not isinstance(target_hook_config, dict):
            message = f"Field `tool.hatch.build.targets.{self.plugin_name}.hooks` must be a table"
            raise TypeError(message)

        for hook_name, config in target_hook_config.items():
            if not isinstance(config, dict):
                message = f"Field `tool.hatch.build.targets.{self.plugin_name}.hooks.{hook_name}` must be a table"
                raise TypeError(message)

            hook_config[hook_name] = config

        if not env_var_enabled(BuildEnvVars.NO_HOOKS):
            all_hooks_enabled = env_var_enabled(BuildEnvVars.HOOKS_ENABLE)
            final_hook_config = {
                hook_name: config
                for hook_name, config in hook_config.items()
                if (
                    all_hooks_enabled
                    or config.get("enable-by-default", True)
                    or env_var_enabled(f"{BuildEnvVars.HOOK_ENABLE_PREFIX}{hook_name.upper()}")
                )
            }
        else:
            final_hook_config = {}

        return final_hook_config

    @cached_property
    def directory(self) -> str:
        if "directory" in self.target_config:
            directory = self.target_config["directory"]
            if not isinstance(directory, str):
                message = f"Field `tool.hatch.build.targets.{self.plugin_name}.directory` must be a string"
                raise TypeError(message)
        else:
            directory = self.build_config.get("directory", DEFAULT_BUILD_DIRECTORY)
            if not isinstance(directory, str):
                message = "Field `tool.hatch.build.directory` must be a string"
                raise TypeError(message)

        return self.normalize_build_directory(directory)

    @cached_property
    def skip_excluded_dirs(self) -> bool:
        if "skip-excluded-dirs" in self.target_config:
            skip_excluded_dirs = self.target_config["skip-excluded-dirs"]
            if not isinstance(skip_excluded_dirs, bool):
                message = f"Field `tool.hatch.build.targets.{self.plugin_name}.skip-excluded-dirs` must be a boolean"
                raise TypeError(message)
        else:
            skip_excluded_dirs = self.build_config.get("skip-excluded-dirs", False)
            if not isinstance(skip_excluded_dirs, bool):
                message = "Field `tool.hatch.build.skip-excluded-dirs` must be a boolean"
                raise TypeError(message)

        return skip_excluded_dirs

    @cached_property
    def ignore_vcs(self) -> bool:
        if "ignore-vcs" in self.target_config:
            ignore_vcs = self.target_config["ignore-vcs"]
            if not isinstance(ignore_vcs, bool):
                message = f"Field `tool.hatch.build.targets.{self.plugin_name}.ignore-vcs` must be a boolean"
                raise TypeError(message)
        else:
            ignore_vcs = self.build_config.get("ignore-vcs", False)
            if not isinstance(ignore_vcs, bool):
                message = "Field `tool.hatch.build.ignore-vcs` must be a boolean"
                raise TypeError(message)

        return ignore_vcs

    @cached_property
    def require_runtime_dependencies(self) -> bool:
        if "require-runtime-dependencies" in self.target_config:
            require_runtime_dependencies = self.target_config["require-runtime-dependencies"]
            if not isinstance(require_runtime_dependencies, bool):
                message = (
                    f"Field `tool.hatch.build.targets.{self.plugin_name}.require-runtime-dependencies` "
                    f"must be a boolean"
                )
                raise TypeError(message)
        else:
            require_runtime_dependencies = self.build_config.get("require-runtime-dependencies", False)
            if not isinstance(require_runtime_dependencies, bool):
                message = "Field `tool.hatch.build.require-runtime-dependencies` must be a boolean"
                raise TypeError(message)

        return require_runtime_dependencies

    @cached_property
    def require_runtime_features(self) -> list[str]:
        if "require-runtime-features" in self.target_config:
            features_config = self.target_config
            features_location = f"tool.hatch.build.targets.{self.plugin_name}.require-runtime-features"
        else:
            features_config = self.build_config
            features_location = "tool.hatch.build.require-runtime-features"

        require_runtime_features = features_config.get("require-runtime-features", [])
        if not isinstance(require_runtime_features, list):
            message = f"Field `{features_location}` must be an array"
            raise TypeError(message)

        all_features: dict[str, None] = {}
        for i, raw_feature in enumerate(require_runtime_features, 1):
            if not isinstance(raw_feature, str):
                message = f"Feature #{i} of field `{features_location}` must be a string"
                raise TypeError(message)

            if not raw_feature:
                message = f"Feature #{i} of field `{features_location}` cannot be an empty string"
                raise ValueError(message)

            feature = normalize_project_name(raw_feature)
            if feature not in self.builder.metadata.core.optional_dependencies:
                message = (
                    f"Feature `{feature}` of field `{features_location}` is not defined in "
                    f"field `project.optional-dependencies`"
                )
                raise ValueError(message)

            all_features[feature] = None

        return list(all_features)

    @cached_property
    def only_packages(self) -> bool:
        """
        Whether or not the target should ignore non-artifact files that do not reside within a Python package.
        """
        if "only-packages" in self.target_config:
            only_packages = self.target_config["only-packages"]
            if not isinstance(only_packages, bool):
                message = f"Field `tool.hatch.build.targets.{self.plugin_name}.only-packages` must be a boolean"
                raise TypeError(message)
        else:
            only_packages = self.build_config.get("only-packages", False)
            if not isinstance(only_packages, bool):
                message = "Field `tool.hatch.build.only-packages` must be a boolean"
                raise TypeError(message)

        return only_packages

    @cached_property
    def reproducible(self) -> bool:
        """
        Whether or not the target should be built in a reproducible manner, defaulting to true.
        """
        if "reproducible" in self.target_config:
            reproducible = self.target_config["reproducible"]
            if not isinstance(reproducible, bool):
                message = f"Field `tool.hatch.build.targets.{self.plugin_name}.reproducible` must be a boolean"
                raise TypeError(message)
        else:
            reproducible = self.build_config.get("reproducible", True)
            if not isinstance(reproducible, bool):
                message = "Field `tool.hatch.build.reproducible` must be a boolean"
                raise TypeError(message)

        return reproducible

    @cached_property
    def dev_mode_dirs(self) -> list[str]:
        """
        Directories which must be added to Python's search path in
        [dev mode](../config/environment/overview.md#dev-mode).
        """
        if "dev-mode-dirs" in self.target_config:
            dev_mode_dirs_config = self.target_config
            dev_mode_dirs_location = f"tool.hatch.build.targets.{self.plugin_name}.dev-mode-dirs"
        else:
            dev_mode_dirs_config = self.build_config
            dev_mode_dirs_location = "tool.hatch.build.dev-mode-dirs"

        all_dev_mode_dirs = []

        dev_mode_dirs = dev_mode_dirs_config.get("dev-mode-dirs", [])
        if not isinstance(dev_mode_dirs, list):
            message = f"Field `{dev_mode_dirs_location}` must be an array of strings"
            raise TypeError(message)

        for i, dev_mode_dir in enumerate(dev_mode_dirs, 1):
            if not isinstance(dev_mode_dir, str):
                message = f"Directory #{i} in field `{dev_mode_dirs_location}` must be a string"
                raise TypeError(message)

            if not dev_mode_dir:
                message = f"Directory #{i} in field `{dev_mode_dirs_location}` cannot be an empty string"
                raise ValueError(message)

            all_dev_mode_dirs.append(dev_mode_dir)

        return all_dev_mode_dirs

    @cached_property
    def dev_mode_exact(self) -> bool:
        if "dev-mode-exact" in self.target_config:
            dev_mode_exact = self.target_config["dev-mode-exact"]
            if not isinstance(dev_mode_exact, bool):
                message = f"Field `tool.hatch.build.targets.{self.plugin_name}.dev-mode-exact` must be a boolean"
                raise TypeError(message)
        else:
            dev_mode_exact = self.build_config.get("dev-mode-exact", False)
            if not isinstance(dev_mode_exact, bool):
                message = "Field `tool.hatch.build.dev-mode-exact` must be a boolean"
                raise TypeError(message)

        return dev_mode_exact

    @cached_property
    def versions(self) -> list[str]:
        # Used as an ordered set
        all_versions: dict[str, None] = {}

        versions = self.target_config.get("versions", [])
        if not isinstance(versions, list):
            message = f"Field `tool.hatch.build.targets.{self.plugin_name}.versions` must be an array of strings"
            raise TypeError(message)

        for i, version in enumerate(versions, 1):
            if not isinstance(version, str):
                message = (
                    f"Version #{i} in field `tool.hatch.build.targets.{self.plugin_name}.versions` must be a string"
                )
                raise TypeError(message)

            if not version:
                message = (
                    f"Version #{i} in field `tool.hatch.build.targets.{self.plugin_name}.versions` "
                    f"cannot be an empty string"
                )
                raise ValueError(message)

            all_versions[version] = None

        if not all_versions:
            default_versions = self.__builder.get_default_versions()
            for version in default_versions:
                all_versions[version] = None
        else:
            unknown_versions = set(all_versions) - set(self.__builder.get_version_api())
            if unknown_versions:
                message = (
                    f"Unknown versions in field `tool.hatch.build.targets.{self.plugin_name}.versions`: "
                    f"{', '.join(map(str, sorted(unknown_versions)))}"
                )
                raise ValueError(message)

        return list(all_versions)

    @cached_property
    def dependencies(self) -> list[str]:
        # Used as an ordered set
        dependencies: dict[str, None] = {}

        target_dependencies = self.target_config.get("dependencies", [])
        if not isinstance(target_dependencies, list):
            message = f"Field `tool.hatch.build.targets.{self.plugin_name}.dependencies` must be an array"
            raise TypeError(message)

        for i, dependency in enumerate(target_dependencies, 1):
            if not isinstance(dependency, str):
                message = (
                    f"Dependency #{i} of field `tool.hatch.build.targets.{self.plugin_name}.dependencies` "
                    f"must be a string"
                )
                raise TypeError(message)

            dependencies[dependency] = None

        global_dependencies = self.build_config.get("dependencies", [])
        if not isinstance(global_dependencies, list):
            message = "Field `tool.hatch.build.dependencies` must be an array"
            raise TypeError(message)

        for i, dependency in enumerate(global_dependencies, 1):
            if not isinstance(dependency, str):
                message = f"Dependency #{i} of field `tool.hatch.build.dependencies` must be a string"
                raise TypeError(message)

            dependencies[dependency] = None

        require_runtime_dependencies = self.require_runtime_dependencies
        require_runtime_features = dict.fromkeys(self.require_runtime_features)
        for hook_name, config in self.hook_config.items():
            hook_require_runtime_dependencies = config.get("require-runtime-dependencies", False)
            if not isinstance(hook_require_runtime_dependencies, bool):
                message = f"Option `require-runtime-dependencies` of build hook `{hook_name}` must be a boolean"
                raise TypeError(message)

            if hook_require_runtime_dependencies:
                require_runtime_dependencies = True

            hook_require_runtime_features = config.get("require-runtime-features", [])
            if not isinstance(hook_require_runtime_features, list):
                message = f"Option `require-runtime-features` of build hook `{hook_name}` must be an array"
                raise TypeError(message)

            for i, raw_feature in enumerate(hook_require_runtime_features, 1):
                if not isinstance(raw_feature, str):
                    message = (
                        f"Feature #{i} of option `require-runtime-features` of build hook `{hook_name}` "
                        f"must be a string"
                    )
                    raise TypeError(message)

                if not raw_feature:
                    message = (
                        f"Feature #{i} of option `require-runtime-features` of build hook `{hook_name}` "
                        f"cannot be an empty string"
                    )
                    raise ValueError(message)

                feature = normalize_project_name(raw_feature)
                if feature not in self.builder.metadata.core.optional_dependencies:
                    message = (
                        f"Feature `{feature}` of option `require-runtime-features` of build hook `{hook_name}` "
                        f"is not defined in field `project.optional-dependencies`"
                    )
                    raise ValueError(message)

                require_runtime_features[feature] = None

            hook_dependencies = config.get("dependencies", [])
            if not isinstance(hook_dependencies, list):
                message = f"Option `dependencies` of build hook `{hook_name}` must be an array"
                raise TypeError(message)

            for i, dependency in enumerate(hook_dependencies, 1):
                if not isinstance(dependency, str):
                    message = f"Dependency #{i} of option `dependencies` of build hook `{hook_name}` must be a string"
                    raise TypeError(message)

                dependencies[dependency] = None

        if require_runtime_dependencies:
            for dependency in self.builder.metadata.core.dependencies:
                dependencies[dependency] = None

        if require_runtime_features:
            for feature in require_runtime_features:
                for dependency in self.builder.metadata.core.optional_dependencies[feature]:
                    dependencies[dependency] = None

        for dependency in self.dynamic_dependencies:
            dependencies[dependency] = None

        return list(dependencies)

    @cached_property
    def dynamic_dependencies(self) -> list[str]:
        dependencies = []
        for hook_name, config in self.hook_config.items():
            build_hook_cls = self.builder.plugin_manager.build_hook.get(hook_name)
            if build_hook_cls is None:
                continue

            # Hook exists but dynamic dependencies are not imported lazily.
            # This happens for example when using the `custom` build hook.
            try:
                build_hook = build_hook_cls(
                    self.root, config, self, self.builder.metadata, "", self.builder.PLUGIN_NAME, self.builder.app
                )
            except ImportError:
                continue

            dependencies.extend(build_hook.dependencies())

        return dependencies

    @cached_property
    def sources(self) -> dict[str, str]:
        if "sources" in self.target_config:
            sources_config = self.target_config
            sources_location = f"tool.hatch.build.targets.{self.plugin_name}.sources"
        else:
            sources_config = self.build_config
            sources_location = "tool.hatch.build.sources"

        sources = {}

        raw_sources = sources_config.get("sources", [])
        if isinstance(raw_sources, list):
            for i, source in enumerate(raw_sources, 1):
                if not isinstance(source, str):
                    message = f"Source #{i} in field `{sources_location}` must be a string"
                    raise TypeError(message)

                if not source:
                    message = f"Source #{i} in field `{sources_location}` cannot be an empty string"
                    raise ValueError(message)

                sources[normalize_relative_directory(source)] = ""
        elif isinstance(raw_sources, dict):
            for source, path in raw_sources.items():
                if not isinstance(path, str):
                    message = f"Path for source `{source}` in field `{sources_location}` must be a string"
                    raise TypeError(message)

                normalized_path = normalize_relative_path(path)
                if normalized_path == ".":
                    normalized_path = ""
                else:
                    normalized_path += os.sep

                sources[normalize_relative_directory(source) if source else source] = normalized_path
        else:
            message = f"Field `{sources_location}` must be a mapping or array of strings"
            raise TypeError(message)

        for relative_path in self.packages:
            source, _package = os.path.split(relative_path)
            if source and normalize_relative_directory(relative_path) not in sources:
                sources[normalize_relative_directory(source)] = ""

        return dict(sorted(sources.items()))

    @cached_property
    def packages(self) -> list[str]:
        if "packages" in self.target_config:
            package_config = self.target_config
            package_location = f"tool.hatch.build.targets.{self.plugin_name}.packages"
        else:
            package_config = self.build_config
            package_location = "tool.hatch.build.packages"

        packages = package_config.get("packages", self.default_packages())
        if not isinstance(packages, list):
            message = f"Field `{package_location}` must be an array of strings"
            raise TypeError(message)

        for i, package in enumerate(packages, 1):
            if not isinstance(package, str):
                message = f"Package #{i} in field `{package_location}` must be a string"
                raise TypeError(message)

            if not package:
                message = f"Package #{i} in field `{package_location}` cannot be an empty string"
                raise ValueError(message)

        return sorted(normalize_relative_path(package) for package in packages)

    @cached_property
    def force_include(self) -> dict[str, str]:
        if "force-include" in self.target_config:
            force_include_config = self.target_config
            force_include_location = f"tool.hatch.build.targets.{self.plugin_name}.force-include"
        else:
            force_include_config = self.build_config
            force_include_location = "tool.hatch.build.force-include"

        force_include = force_include_config.get("force-include", {})
        if not isinstance(force_include, dict):
            message = f"Field `{force_include_location}` must be a mapping"
            raise TypeError(message)

        for i, (source, relative_path) in enumerate(force_include.items(), 1):
            if not source:
                message = f"Source #{i} in field `{force_include_location}` cannot be an empty string"
                raise ValueError(message)

            if not isinstance(relative_path, str):
                message = f"Path for source `{source}` in field `{force_include_location}` must be a string"
                raise TypeError(message)

            if not relative_path:
                message = f"Path for source `{source}` in field `{force_include_location}` cannot be an empty string"
                raise ValueError(message)

        return normalize_inclusion_map(force_include, self.root)

    @cached_property
    def only_include(self) -> dict[str, str]:
        if "only-include" in self.target_config:
            only_include_config = self.target_config
            only_include_location = f"tool.hatch.build.targets.{self.plugin_name}.only-include"
        else:
            only_include_config = self.build_config
            only_include_location = "tool.hatch.build.only-include"

        only_include = only_include_config.get("only-include", self.default_only_include()) or self.packages
        if not isinstance(only_include, list):
            message = f"Field `{only_include_location}` must be an array"
            raise TypeError(message)

        inclusion_map = {}

        for i, relative_path in enumerate(only_include, 1):
            if not isinstance(relative_path, str):
                message = f"Path #{i} in field `{only_include_location}` must be a string"
                raise TypeError(message)

            normalized_path = normalize_relative_path(relative_path)
            if not normalized_path or normalized_path.startswith(("~", "..")):
                message = f"Path #{i} in field `{only_include_location}` must be relative: {relative_path}"
                raise ValueError(message)

            if normalized_path in inclusion_map:
                message = f"Duplicate path in field `{only_include_location}`: {normalized_path}"
                raise ValueError(message)

            inclusion_map[normalized_path] = normalized_path

        return normalize_inclusion_map(inclusion_map, self.root)

    def get_distribution_path(self, relative_path: str) -> str:
        # src/foo/bar.py -> foo/bar.py
        for source, replacement in self.sources.items():
            if not source:
                return replacement + relative_path

            if relative_path.startswith(source):
                return relative_path.replace(source, replacement, 1)

        return relative_path

    @cached_property
    def vcs_exclusion_files(self) -> dict[str, list[str]]:
        exclusion_files: dict[str, list[str]] = {"git": [], "hg": []}

        local_gitignore = locate_file(self.root, ".gitignore", boundary=".git")
        if local_gitignore is not None:
            exclusion_files["git"].append(local_gitignore)

        local_hgignore = locate_file(self.root, ".hgignore", boundary=".hg")
        if local_hgignore is not None:
            exclusion_files["hg"].append(local_hgignore)

        return exclusion_files

    def load_vcs_exclusion_patterns(self) -> list[str]:
        patterns = []

        # https://git-scm.com/docs/gitignore#_pattern_format
        for exclusion_file in self.vcs_exclusion_files["git"]:
            with open(exclusion_file, encoding="utf-8") as f:
                patterns.extend(f.readlines())

        # https://linux.die.net/man/5/hgignore
        for exclusion_file in self.vcs_exclusion_files["hg"]:
            with open(exclusion_file, encoding="utf-8") as f:
                glob_mode = False
                for line in f:
                    exact_line = line.strip()
                    if exact_line == "syntax: glob":
                        glob_mode = True
                        continue

                    if exact_line.startswith("syntax: "):
                        glob_mode = False
                        continue

                    if glob_mode:
                        patterns.append(line)

        # validate project root is not excluded by vcs
        exclude_spec = pathspec.GitIgnoreSpec.from_lines(patterns)
        if exclude_spec.match_file(self.root):
            return []

        return patterns

    def normalize_build_directory(self, build_directory: str) -> str:
        if not os.path.isabs(build_directory):
            build_directory = os.path.join(self.root, build_directory)

        return os.path.normpath(build_directory)

    def default_include(self) -> list:  # noqa: PLR6301
        return []

    def default_exclude(self) -> list:  # noqa: PLR6301
        return []

    def default_packages(self) -> list:  # noqa: PLR6301
        return []

    def default_only_include(self) -> list:  # noqa: PLR6301
        return []

    def default_global_exclude(self) -> list[str]:  # noqa: PLR6301
        patterns = ["*.py[cdo]", f"/{DEFAULT_BUILD_DIRECTORY}"]
        patterns.sort()
        return patterns

    def set_exclude_all(self) -> None:
        self.__exclude_all = True

    def get_force_include(self) -> dict[str, str]:
        force_include = self.force_include.copy()
        force_include.update(self.build_force_include)
        return force_include

    @contextmanager
    def set_build_data(self, build_data: dict[str, Any]) -> Generator:
        try:
            # Include anything the hooks indicate
            build_artifacts = build_data["artifacts"]
            if build_artifacts:
                self.build_artifact_spec = pathspec.GitIgnoreSpec.from_lines(build_artifacts)

            self.build_force_include.update(normalize_inclusion_map(build_data["force_include"], self.root))

            for inclusion_map in (self.force_include, self.build_force_include):
                for source, target in inclusion_map.items():
                    # Ignore source
                    # old/ -> new/
                    # old.ext -> new.ext
                    if source.startswith(f"{self.root}{os.sep}"):
                        self.build_reserved_paths.add(self.get_distribution_path(os.path.relpath(source, self.root)))
                    # Ignore target files only
                    # ../out.ext -> ../in.ext
                    elif os.path.isfile(source):
                        self.build_reserved_paths.add(self.get_distribution_path(target))

            yield
        finally:
            self.build_artifact_spec = None
            self.build_force_include.clear()
            self.build_reserved_paths.clear()


def env_var_enabled(env_var: str, *, default: bool = False) -> bool:
    if env_var in os.environ:
        return os.environ[env_var] in {"1", "true"}

    return default


BuilderConfigBound = TypeVar("BuilderConfigBound", bound=BuilderConfig)
